import finesse
import numpy as np
from finesse.analysis.actions import Action, Series, RunLocks, OptimiseRFReadoutPhaseDC

def make_aligo(RF_AS_readout=False, verbose=False):
    base = finesse.Model()
    base.parse(aligo_katscript)

    OptimiseRFReadoutPhaseDC(
        "CARM", "REFL9",
        "PRCL", "POP9",
        "SRCL", "POP45",
        "DARM", "AS45",
    ).run(base)

    set_lock_gains(base, verbose=verbose)

    if not RF_AS_readout:
        DARM_RF_to_DC().run(base)

    return base

class DARM_RF_to_DC(Action):
    """Locks a model using DARM RF readout then transitions the model into
    using a DC readout and locks.
    """
    def __init__(self, name="DarmRF2DC"):
        super().__init__(name)
        self.__lock_rf = RunLocks('DARM_rf_lock')
        self.__lock_dc = RunLocks('DARM_dc_lock')
        
    def _do(self, state):
        self.__lock_rf._do(state)
        state.model.DARM_rf_lock.disabled = True
        # kick lock away from zero tuning for DC lock to grab with
        state.model.DARM.DC += 0.5e-3 
        # take a guess at the gain
        state.model.DARM_dc_lock.gain = -0.01 
        state.model.DARM_dc_lock.disabled = False
        self.__lock_dc._do(state)
        return None
        
    def _requests(self, model, memo, first=True):
        self.__lock_rf._requests(model, memo)
        self.__lock_dc._requests(model, memo)
        return memo
    

class DRFPMI_state(Action):
    """Assumes a mode has a PRM, SRM, ITMX, ETMX, ITMY, and ETMY mirror elements
    in. This action will change the alignment state of these. The options are:
    
    'PRMI', 'SRMI', 'MI', 'FPMI', 'PRFPMI', 'SRFPMI', 'DRFPMI', 'XARM', 'YARM'
    
    This action will change the state of the model.
    """
    def __init__(self, state : str, name="drfpmi_state"):
        super().__init__(name)
        states = ('PRMI', 'SRMI', 'MI', 'FPMI', 'PRFPMI', 'SRFPMI', 'DRFPMI', 'XARM', 'YARM')
        if state not in states:
            raise ValueError(f"State '{state}' is not a valid option: {states}")
        self.state = state
        
    def _do(self, state):
        if self.state == 'PRMI':
            state.model.PRM.misaligned = 0
            state.model.SRM.misaligned = 1
            state.model.ETMX.misaligned = 1
            state.model.ITMX.misaligned = 0
            state.model.ETMY.misaligned = 1
            state.model.ITMY.misaligned = 0
        elif self.state == 'SRMI':
            state.model.PRM.misaligned = 1
            state.model.SRM.misaligned = 0
            state.model.ETMX.misaligned = 1
            state.model.ITMX.misaligned = 0
            state.model.ETMY.misaligned = 1
            state.model.ITMY.misaligned = 0
        elif self.state == 'MI':
            state.model.PRM.misaligned = 1
            state.model.SRM.misaligned = 1
            state.model.ETMX.misaligned = 1
            state.model.ITMX.misaligned = 0
            state.model.ETMY.misaligned = 1
            state.model.ITMY.misaligned = 0
        elif self.state == 'FPMI':
            state.model.PRM.misaligned = 1
            state.model.SRM.misaligned = 1
            state.model.ETMX.misaligned = 0
            state.model.ITMX.misaligned = 0
            state.model.ETMY.misaligned = 0
            state.model.ITMY.misaligned = 0
        elif self.state == 'PRFPMI':
            state.model.PRM.misaligned = 0
            state.model.SRM.misaligned = 1
            state.model.ETMX.misaligned = 0
            state.model.ITMX.misaligned = 0
            state.model.ETMY.misaligned = 0
            state.model.ITMY.misaligned = 0
        elif self.state == 'SRFPMI':
            state.model.PRM.misaligned = 1
            state.model.SRM.misaligned = 0
            state.model.ETMX.misaligned = 0
            state.model.ITMX.misaligned = 0
            state.model.ETMY.misaligned = 0
            state.model.ITMY.misaligned = 0
        elif self.state == 'DRFPMI':
            state.model.PRM.misaligned = 0
            state.model.SRM.misaligned = 0
            state.model.ETMX.misaligned = 0
            state.model.ITMX.misaligned = 0
            state.model.ETMY.misaligned = 0
            state.model.ITMY.misaligned = 0
        elif self.state == 'YARM':
            state.model.PRM.misaligned = 1
            state.model.SRM.misaligned = 1
            state.model.ETMX.misaligned = 1
            state.model.ITMX.misaligned = 1
            state.model.ETMY.misaligned = 0
            state.model.ITMY.misaligned = 0
        elif self.state == 'XARM':
            state.model.PRM.misaligned = 1
            state.model.SRM.misaligned = 1
            state.model.ETMX.misaligned = 0
            state.model.ITMX.misaligned = 0
            state.model.ETMY.misaligned = 1
            state.model.ITMY.misaligned = 1
        else:
            raise Exception(f"{self.state} not implemented")
        
    def _requests(self, model, memo, first=True):
        # changing the mirror misaligned parameter is essentially
        # changing the mirror reflectivity model parameter
        memo['changing_parameters'].extend((
            "PRM.misaligned",
            "SRM.misaligned",
            "ETMX.misaligned",
            "ITMX.misaligned",
            "ETMY.misaligned",
            "ITMY.misaligned")
        )
        return memo


def set_lock_gains(model, d_dof=1e-6, gain_scale=1, verbose=False):
    """For the current state of the model each lock will have its
    gain computed. This is done by computing the gradient of the
    error signal with respect to the set feedback.
    
    The optical gain is then computed as -1/(slope).
    
    This function alters the state of the provided model.
    
    Parameters
    ----------
    model : Model
        Model to set the lock gains of
    d_dof : double
        step size for computing the slope of the error signals
    verbose : boolean
        Prints information when true
    """
    from finesse.analysis.actions import Xaxis, Series
    from finesse.components.readout import ReadoutOutput
    
    for lock in model.locks:
        # Make sure readouts being used have their outputs enabled
        if type(lock.error_signal) is ReadoutOutput:
            lock.error_signal.readout.output_detectors = True
            
    # Use a flattened series analysis as it only creates one model
    # and xaxis resets all the parameters each time
    analysis = Series(*(
            Xaxis(lock.feedback, 'lin', -d_dof, d_dof, 1, relative=True, name=lock.name) for lock in model.locks
        ),
        flatten=True 
    )
    sol = analysis.run(model)
    N = len(model.locks)
    for lock in model.locks:
        lock_sol = sol[lock.name]
        x = lock_sol.x1
        error = lock_sol[lock.error_signal.name] + lock.offset
        grad = np.gradient(error, x[1]-x[0]).mean()
        if grad == 0:
            lock.gain = np.NaN
        else:
            lock.gain = -1/grad * gain_scale
    
        if verbose:
            print(lock, lock.error_signal.name, lock.gain)


def get_lock_error_signals(model, dof_range, steps=1000, verbose=False):
    from finesse.analysis.actions import Xaxis, Series
    from finesse.components.readout import ReadoutOutput
    
    for lock in model.locks:
        # Make sure readouts being used have their outputs enabled
        if type(lock.error_signal) is ReadoutOutput:
            lock.error_signal.readout.output_detectors = True
            
    # Use a flattened series analysis as it only creates one model
    # and xaxis resets all the parameters each time
    analysis = Series(*(
            Xaxis(lock.feedback, 'lin', -dof_range, dof_range, steps, relative=True, name=lock.feedback.owner.name) for lock in model.locks
        ),
        flatten=True 
    )
    sol = analysis.run(model)
    return sol


aligo_katscript = """
# modulators for core interferometer sensing - Advanced LIGO, CQG, 2015
# http://iopscience.iop.org/article/10.1088/0264-9381/32/7/074001/meta#cqg507871s4-8
# 9MHz (CARM, PRC, SRC loops)
variable f1 9099471
variable f2 5*&f1
variable nsilica 1.45
variable Mloss 30u
###############################################################################
###   length definitions
###############################################################################
variable Larm 3994
variable LPR23 16.164  # distance between PR2 and PR3
variable LSR23 15.443  # distance between SR2 and SR3
variable LPR3BS 19.538 # distance between PR3 and BS
variable LSR3BS 19.366 # distance between SR3 and BS
variable lmich 5.342   # average length of MICH
variable lschnupp 0.08
variable lPRC (3+0.5)*c0/(2*&f1) # T1000298 Eq2.1, N=3
variable lSRC (17)*c0/(2*&f2) # T1000298 Eq2.2, M=3
###############################################################################
###   laser
###############################################################################
laser L0 P=125
mod mod1 f=&f1 midx=0.18 order=1 mod_type=pm
mod mod2 f=&f2 midx=0.18 order=1 mod_type=pm
link(L0, mod1, mod2)
###############################################################################
###   PRC
###############################################################################
s sPRCin mod2.p2 PRMAR.p1
m PRMAR R=0 L=40u xbeta=&PRM.xbeta ybeta=&PRM.ybeta phi=&PRM.phi
s sPRMsub1 PRMAR.p2 PRM.p1 L=0.0737 nr=&nsilica
m PRM T=0.03 L=8.5u Rc=11.009
s lp1 PRM.p2 PR2.p1 L=&lPRC-&LPR3BS-&LPR23-&lmich
bs PR2 T=250u L=&Mloss alpha=-0.79 Rc=-4.545
s lp2 PR2.p2 PR3.p1 L=&LPR23
bs PR3 T=0 L=&Mloss alpha=0.615 Rc=36.027
s lp3 PR3.p2 BS.p1 L=&LPR3BS
###############################################################################
###   BS
###############################################################################
bs BS R=0.5 L=&Mloss alpha=45
s BSsub1 BS.p3 BSAR1.p1 L=0.0687 nr=&nsilica
s BSsub2 BS.p4 BSAR2.p2 L=0.0687 nr=&nsilica
bs BSAR1 L=50u R=0 alpha=-29.195
bs BSAR2 L=50u R=0 alpha=29.195
###############################################################################
###   Yarm
###############################################################################
# Distance from beam splitter to Y arm input mirror
s ly1 BS.p2 ITMYlens.p1 L=&lmich-&lschnupp/2-&ITMYsub.L*&ITMXsub.nr
lens ITMYlens f=34500
s ly2 ITMYlens.p2 ITMYAR.p1
m ITMYAR R=0 L=20u xbeta=&ITMY.xbeta ybeta=&ITMY.ybeta phi=&ITMY.phi
s ITMYsub ITMYAR.p2 ITMY.p1 L=0.2 nr=&nsilica
m ITMY T=0.014 L=&Mloss Rc=-1934
s LY ITMY.p2 ETMY.p1 L=&Larm
m ETMY T=5u L=&Mloss Rc=2245
s ETMYsub ETMY.p2 ETMYAR.p1 L=0.2 nr=&nsilica
m ETMYAR 0 500u xbeta=&ETMY.xbeta ybeta=&ETMY.ybeta phi=&ETMY.phi
###############################################################################
###   Xarm
###############################################################################
# Distance from beam splitter to X arm input mirror
s lx1 BSAR1.p3 ITMXlens.p1 L=&lmich+&lschnupp/2-&ITMXsub.L*&ITMXsub.nr-&BSsub1.L*&BSsub1.nr
lens ITMXlens f=34500
s lx2 ITMXlens.p2 ITMXAR.p1
m ITMXAR R=0 L=20u xbeta=&ITMX.xbeta ybeta=&ITMX.ybeta phi=&ITMX.phi
s ITMXsub ITMXAR.p2 ITMX.p1 L=0.2 nr=&nsilica
m ITMX T=0.014 L=&Mloss Rc=-1934
s LX ITMX.p2 ETMX.p1 L=&Larm
m ETMX T=5u L=&Mloss Rc=2245
s ETMXsub ETMX.p2 ETMXAR.p1 L=0.2 nr=&nsilica
m ETMXAR 0 500u xbeta=&ETMX.xbeta ybeta=&ETMX.ybeta phi=&ETMX.phi
###############################################################################
###   SRC
###############################################################################
s ls3 BSAR2.p4 SR3.p1 L=&LSR3BS
bs SR3 T=0 L=&Mloss alpha=0.785 Rc=35.972841
s ls2 SR3.p2 SR2.p1 L=&LSR23
bs SR2 T=0 L=&Mloss alpha=-0.87 Rc=-6.406
s ls1 SR2.p2 SRM.p1 L=&lSRC-&LSR3BS-&LSR23-&BSsub2.L*&BSsub2.nr-&lmich  
m SRM T=0.32 L=8.7u Rc=-5.6938
s SRMsub SRM.p2 SRMAR.p1 L=0.0749 nr=&nsilica
m SRMAR R=0 L=50n
###############################################################################
###   OMC
###############################################################################
# Here we just use some simple filter to approximate an OMC for filtering
# out RF fields, this doesn't filter HOMs!
obp OMC fc=0 bandwidth=1M
link(SRMAR.p2, OMC)
###############################################################################
### Length sensing and control
###############################################################################
dof XARM ETMX.dofs.z
dof YARM ETMY.dofs.z
dof CARM ETMX.dofs.z +1 ETMY.dofs.z +1
dof DARM ETMX.dofs.z +1 ETMY.dofs.z -1
dof PRCL PRM.dofs.z +1 
dof SRCL SRM.dofs.z +1 DC=90
dof MICH BS.dofs.z +1
dof MICH2 ITMY.dofs.z +1 ETMY.dofs.z +1 ITMX.dofs.z -1 ETMX.dofs.z -1
dof STRAIN LX.dofs.h +1 LY.dofs.h -1
dof FRQ L0.dofs.frq
dof RIN L0.dofs.amp

readout_rf REFL9 PRMAR.p1.o f=&f1
readout_rf REFL18 PRMAR.p1.o f=3*&f1
readout_rf REFL45 PRMAR.p1.o f=5*&f1
readout_rf POP9  PR2.p3.o   f=&f1
readout_rf POP45 PR2.p3.o   f=&f2
readout_rf AS45  SRMAR.p2.o f=&f2
readout_dc AS    OMC.p2.o

lock CARM_lock REFL9.outputs.I CARM.DC -0.1 1e-6 
lock MICH_lock POP45.outputs.Q MICH.DC -15 1e-6
lock PRCL_lock POP9.outputs.I PRCL.DC 2.8 1e-6   
lock SRCL_lock POP45.outputs.I SRCL.DC 42 1e-6 
lock DARM_rf_lock AS45.outputs.I DARM.DC -0.003 1e-6
lock DARM_dc_lock AS.outputs.DC DARM.DC -0.003 1e-6 offset=20m disabled=true
###############################################################################
### DC power measurements
###############################################################################
pd Px ETMX.p1.i   
pd Py ETMX.p1.i   
pd Pprc PRM.p2.o  
pd Psrc SRM.p1.i  
pd Prefl ETMX.p1.i
pd Pas OMC.p2.o
"""



import pytest
import finesse


@pytest.fixture
def kat():
    kat = finesse.Model()
    kat.parse(
        """
        l l1 P=1
        s s1 l1.p1 m1.p1
        m m1 R=1 T=0
        free_mass m1_sus m1.mech mass=1

        sgen sig l1.amp.i 1 0

        ad z m1.mech.z &fsig.f
        ad F_z m1.mech.F_z &fsig.f

        fsig(1)
        """
    )
    return kat


def test_amplitude(kat):
    sol = kat.run()
    from scipy.constants import c, pi

    P = kat.l1.P.value
    M = kat.m1_sus.mass.value
    w = 2 * pi * kat.fsig.f.value
    analytic = -2 * P / c * 1 / (-M * w ** 2)
    assert abs(sol["z"] - analytic) < 1e-15



"""
Computes the laser power in an interferometer output or the power from an electrical signal.
"""

import logging
import numbers

import numpy as np
import finesse.detectors._pdtypes as pdtypes
from finesse.detectors.compute import (
    pd0_DC_output,
    pd0_DC_output_segmented,
    pd0_DC_output_masked,
    pd1_DC_output,
    pd1_AC_output,
    pd2_DC_output,
    pd2_AC_output,
)
from finesse.detectors.compute.power import PD0Workspace, PD1Workspace, PD2Workspace
from finesse.detectors.general import MaskedDetector
from finesse.parameter import float_parameter, ParameterState, Parameter


LOGGER = logging.getLogger(__name__)


def check_is_audio(ws, f: Parameter):
    if ws.sim.signal is None:
        return False
    elif f.state == ParameterState.Symbolic:
        return f.value.owner is ws.sim.model.fsig
    else:
        return f.value is ws.sim.model.fsig.f.value


@float_parameter("f", "Frequency")
@float_parameter("phase", "Phase")
class PowerDetectorDemod1(MaskedDetector):
    """Represents a power detector with one RF demodulation.
    It calculates the RF beat power at a node in Watts of optical power.

    If no demodulation phase is specified then this detector outputs a
    complex value `I+1j*Q`.

    Parameters
    ----------
    name : str
        Name of newly created power detector.

    node : :class:`.Node`
        Node to read output from.

    f : float
        Demodulation frequency in Hz

    phase : float, optional
        Demodulation phase in degrees
    """

    def __init__(self, name, node, f, phase=None):
        if f is None:
            raise ValueError("A demodulation frequency must be provided")

        if phase is not None:
            self.__mode = "mixer_real"
            dtype = np.float64
        else:
            self.__mode = "mixer_complex"
            dtype = np.complex128

        self._beats = None

        MaskedDetector.__init__(self, name, node, dtype=dtype, unit="W", label="Power")
        self.f = f
        self.phase = phase

    def _get_workspace(self, sim):
        ws = PD1Workspace(self, sim)

        ws.is_f_changing = self.f.is_changing
        ws.is_phase_changing = self.phase.is_changing

        if ws.is_phase_changing or self.phase.value is not None:
            # We might change from None to some actual value
            ws.output_real = True
            self._set_dtype(np.float64)
        else:
            # If no phase defined output complex power
            ws.output_real = False
            self._set_dtype(np.complex128)

        ws.dc_node_id = sim.carrier.node_id(self.node)
        if sim.signal:
            ws.ac_node_id = sim.signal.node_id(self.node)
            ws.is_audio_mixing = check_is_audio(ws, self.f)

        # Would there be any weird situation where AC and DC homs are different?
        ws.homs = ws.sim.model_data.homs_view

        if ws.is_audio_mixing:
            ws.set_output_fn(pd1_AC_output)
        else:
            ws.set_output_fn(pd1_DC_output)

        if not ws.is_f_changing and not ws.is_audio_mixing:
            # Sidebands beating together are known apriori if frequency bins are not
            # changing, or if this is just an audio mixer.
            ws.update_parameter_values()
            # If frequency is fixed then we just precompute the beats
            ws.update_beats()

        return ws


@float_parameter("f1", "Frequency 1")
@float_parameter("phase1", "Phase 1")
@float_parameter("f2", "Frequency 2")
@float_parameter("phase2", "Phase 2")
class PowerDetectorDemod2(MaskedDetector):
    """Represents a power detector with two RF demodulation.
    It calculates the RF beat power at a node in Watts of optical power.

    If no demodulation phase is specified for the final demodulation
    this detector outputs a complex value `I+1j*Q` where I and Q are
    the in-phase and quadrature parts of the signal.

    Parameters
    ----------
    name : str
        Name of newly created power detector.

    node : :class:`.Node`
        Node to read output from.

    f1 : float
        First demodulation frequency in Hz

    phase1 : float
        First demodulation phase in degrees

    f2 : float
        Second demodulation frequency in Hz

    phase2 : float, optional
        Second demodulation phase in degrees
    """

    def __init__(self, name, node, f1, phase1, f2, phase2=None):
        if phase2 is not None:
            self.__mode = "mixer_real"
            dtype = np.float64
        else:
            self.__mode = "mixer_complex"
            dtype = np.complex128

        self._beats = None

        MaskedDetector.__init__(self, name, node, dtype=dtype, unit="W", label="Power")
        self.f1 = f1
        self.phase1 = phase1
        self.f2 = f2
        self.phase2 = phase2

    def _get_workspace(self, sim):
        ws = PD2Workspace(self, sim)

        ws.is_f1_changing = self.f1.is_changing
        ws.is_phase1_changing = self.phase1.is_changing
        ws.is_f2_changing = self.f2.is_changing
        ws.is_phase2_changing = self.phase2.is_changing

        if ws.is_phase2_changing or self.phase2.value is not None:
            # We might change from None to some actual value
            ws.output_real = True
            self._set_dtype(np.float64)
        else:
            # If no phase defined output complex power
            ws.output_real = False
            self._set_dtype(np.complex128)

        ws.dc_node_id = sim.carrier.node_id(self.node)

        if sim.signal:
            ws.ac_node_id = sim.signal.node_id(self.node)
            if check_is_audio(ws, self.f1):
                raise Exception(
                    f"pd2 {self.name} f1 cannot be an audio frequency, use f2 for audio demodulation"
                )
            ws.is_audio_mixing = check_is_audio(ws, self.f2)

        # Would there be any weird situation where AC and DC homs are different?
        ws.homs = ws.sim.model_data.homs_view

        if ws.is_audio_mixing:
            ws.set_output_fn(pd2_AC_output)
        else:
            ws.set_output_fn(pd2_DC_output)

        # if not ws.is_f1_changing and not (ws.is_f2_changing and ws.is_audio_mixing):
        #     # Sidebands beating together are known apriori if frequency bins are not
        #     # changing, or if this is just an audio mixer.
        #     ws.update_parameter_values()
        #     # If frequency is fixed then we just precompute the beats
        #     ws.update_beats()

        return ws


class PowerDetector(MaskedDetector):
    """Represents a power detector with no RF demodulations.
    It calculates the DC laser power at a node in Watts of optical power.

    Parameters
    ----------
    name : str
        Name of newly created power detector.

    node : :class:`.Node`
        Node to read output from.
    """

    def __init__(self, name, node, *, pdtype=None):
        MaskedDetector.__init__(
            self, name, node, dtype=np.float64, unit="W", label="Power"
        )
        self.pdtype = getattr(pdtypes, pdtype.upper()) if isinstance(pdtype, str) else pdtype

    def _get_workspace(self, sim):
        ws = PD0Workspace(self, sim)

        ni = sim.carrier.get_node_info(self.node)
        ws.rhs_index = ni["rhs_index"]
        ws.size = ni["nfreqs"] * ni["nhoms"]
        ws.dc_node_id = sim.carrier.node_id(self.node)

        if ws.has_mask:
            if self.pdtype: # not supported yet
                raise NotImplementedError()
            ws.set_output_fn(pd0_DC_output_masked)
        else:
            if self.pdtype:
                import finesse.detectors._pdtypes as pdtype
                ws.tmp = np.zeros(ws.size, dtype=complex)
                ws.K = pdtype.construct_segment_beat_matrix(sim.model.mode_index_map, self.pdtype)
                ws.set_output_fn(pd0_DC_output_segmented)
            else:
                ws.set_output_fn(pd0_DC_output)
        return ws


@float_parameter("f1", "Frequency 1")
@float_parameter("f2", "Frequency 2")
@float_parameter("f3", "Frequency 3")
@float_parameter("f4", "Frequency 4")
@float_parameter("f5", "Frequency 5")
@float_parameter("f6", "Frequency 6")
@float_parameter("f7", "Frequency 7")
@float_parameter("f8", "Frequency 8")
@float_parameter("f9", "Frequency 9")
@float_parameter("phase1", "Phase 1")
@float_parameter("phase2", "Phase 2")
@float_parameter("phase3", "Phase 3")
@float_parameter("phase4", "Phase 4")
@float_parameter("phase5", "Phase 5")
@float_parameter("phase6", "Phase 6")
@float_parameter("phase7", "Phase 7")
@float_parameter("phase8", "Phase 8")
@float_parameter("phase9", "Phase 9")
class CustomPD(PowerDetector):
    """A custom power detector with beat coefficients describing the coupling of modes."""

    def __init__(self, name, node, beats, **kwargs):
        PowerDetector.__init__(self, name, node, **kwargs)

        if isinstance(beats, np.ndarray):
            self._beats = beats
            self._beats_dict = None
        elif isinstance(beats, dict):
            # the field _beats will be replaced by a np array once
            # construct beats is called (this happens automatically
            # when adding a CustomPD object to a model and / or
            # changing the maxtem value of the model)
            self._beats = None
            self._beats_dict = beats

    def construct_beats(self, maxtem=None):
        """Constructs, or re-constructs, the beat coefficients matrix
        based on the value of `maxtem`.

        The underlying generic beats dictionary passed during construction
        is used to generate the beat factor matrix. If a beats matrix was
        passed directly during construction of the CustomPD, then this
        method has no effect.

        Parameters
        ----------
        maxtem : int, optional
            The maximum TEM order to generate the beat coefficients up
            to. Defaults to `None` such that this value is taken from
            the model associated with this detector.
        """
        if self._beats_dict is None:
            return

        if maxtem is None:
            maxtem = self._model.modes_setting["maxtem"]

            if maxtem is None:
                return

        self._beats = np.zeros((1 + maxtem, 1 + maxtem, 1 + maxtem, 1 + maxtem))

        for (n1, m1, n2, m2), factor in self._beats_dict.items():
            if all(isinstance(k, numbers.Number) for k in (n1, m1, n2, m2)):
                self._beats[int(n1)][int(m1)][int(n2)][int(m2)] = factor
            else:
                if all(k == "x" for k in (m1, m2)) and all(
                    isinstance(k, numbers.Number) for k in (n1, n2)
                ):
                    n1 = int(n1)
                    n2 = int(n2)
                    if n1 > maxtem or n2 > maxtem:
                        continue

                    for i in range(maxtem):
                        self._beats[n1][i][n2][i] = factor
                        self._beats[n2][i][n1][i] = factor
                elif all(k == "x" for k in (n1, n2)) and all(
                    isinstance(k, numbers.Number) for k in (m1, m2)
                ):
                    m1 = int(m1)
                    m2 = int(m2)
                    if m1 > maxtem or m2 > maxtem:
                        continue

                    for i in range(maxtem):
                        self._beats[i][m1][i][m2] = factor
                        self._beats[i][m2][i][m1] = factor


@float_parameter("f1", "Frequency 1")
@float_parameter("f2", "Frequency 2")
@float_parameter("f3", "Frequency 3")
@float_parameter("f4", "Frequency 4")
@float_parameter("f5", "Frequency 5")
@float_parameter("f6", "Frequency 6")
@float_parameter("f7", "Frequency 7")
@float_parameter("f8", "Frequency 8")
@float_parameter("f9", "Frequency 9")
@float_parameter("phase1", "Phase 1")
@float_parameter("phase2", "Phase 2")
@float_parameter("phase3", "Phase 3")
@float_parameter("phase4", "Phase 4")
@float_parameter("phase5", "Phase 5")
@float_parameter("phase6", "Phase 6")
@float_parameter("phase7", "Phase 7")
@float_parameter("phase8", "Phase 8")
@float_parameter("phase9", "Phase 9")
class SplitPD(CustomPD):
    def __init__(self, name, node, direction, **kwargs):
        if direction == "x":
            beats = pdtypes.XSPLIT
        else:
            beats = pdtypes.YSPLIT

        CustomPD.__init__(self, name, node, beats, **kwargs)

        self.__direction = direction



from finesse.script.containers import (
    KatMetaToken,
    KatToken,
    KatNumberToken,
    KatStringToken,
    KatNoneToken,
    KatReferenceToken,
    KatBooleanToken,
    KatWhitespaceToken,
)

# Tokens.
NEWLINE = lambda lineno, index: KatToken(lineno, index, index + 1, "NEWLINE", "\n")
SPACE = lambda lineno, start, length=1: KatWhitespaceToken(
    lineno, start, start + length, "WHITESPACE", " " * length
)
COMMENT = lambda lineno, start, value: KatToken(
    lineno, start, start + len(value), "COMMENT", value
)
LINEEND = lambda lineno, start: KatToken(lineno, start, start + 1, "NEWLINE", "\n")
NAME = lambda lineno, start, value: KatToken(
    lineno, start, start + len(value), "NAME", value
)
NUMBER = lambda lineno, start, value: KatNumberToken(
    lineno, start, start + len(value), "NUMBER", value
)
STRING = lambda lineno, start, value: KatStringToken(
    lineno, start, start + len(value), "STRING", value
)
NONE = lambda lineno, start: KatNoneToken(
    lineno, start, start + len("none"), "NONE", "none"
)
REFERENCE = lambda lineno, start, value: KatReferenceToken(
    lineno, start, start + len(value), "REFERENCE", value
)
BOOLEAN = lambda lineno, start, value: KatBooleanToken(
    lineno, start, start + len(str(value)), "BOOLEAN", value
)
LBRACKET = lambda lineno, start: KatToken(lineno, start, start + 1, "LBRACKET", "[")
RBRACKET = lambda lineno, start: KatToken(lineno, start, start + 1, "RBRACKET", "]")
LPAREN = lambda lineno, start: KatToken(lineno, start, start + 1, "LPAREN", "(")
RPAREN = lambda lineno, start: KatToken(lineno, start, start + 1, "RPAREN", ")")
COMMA = lambda lineno, start: KatToken(lineno, start, start + 1, "COMMA", ",")
PLUS = lambda lineno, start: KatToken(lineno, start, start + 1, "PLUS", "+")
MINUS = lambda lineno, start: KatToken(lineno, start, start + 1, "MINUS", "-")
TIMES = lambda lineno, start: KatToken(lineno, start, start + 1, "TIMES", "*")
DIVIDE = lambda lineno, start: KatToken(lineno, start, start + 1, "DIVIDE", "/")
FLOORDIVIDE = lambda lineno, start: KatToken(
    lineno, start, start + 2, "FLOORDIVIDE", "//"
)
POWER = lambda lineno, start: KatToken(lineno, start, start + 2, "POWER", "**")
EQUALS = lambda lineno, start: KatToken(lineno, start, start + 1, "EQUALS", "=")

# Meta tokens.
IMPLICITLINEEND = lambda lineno, start: KatMetaToken(lineno, start, start, "NEWLINE")
ENDMARKER = lambda lineno: KatMetaToken(lineno, 1, 1, "ENDMARKER")



"""Accumulated Gouy phase detector."""

import logging
import numpy as np

from finesse.detectors.general import Detector
from finesse.detectors.compute.gaussian import GouyDetectorWorkspace

LOGGER = logging.getLogger(__name__)


class Gouy(Detector):
    """Detector to measure the accumulated gouy phase across a sequence of spaces.

    This detector can operate in one of two modes, depending upon args given:

      * automatically determine the spaces through an arbitrary path by specifying the
        `from_node` and `to_node` arguments,
      * OR provide a pre-determined sequence of spaces (or their names) as positional
        arguments.

    Whichever option is chosen, this detector will compute the same fundamental quantity;
    that is the sum of the Gouy phases accumulated over each space (in degrees).

    Parameters
    ----------
    name : str
        Name of newly created gouy detector.

    *args : sequence
        A sequence of spaces or space names.

    from_node : node-like
        An :class:`.OpticalNode` instance, or a data type which can be converted
        to an optical node.

    to_node : node-like
        An :class:`.OpticalNode` instance, or a data type which can be converted
        to an optical node.

    via_node : node-like
        An :class:`.OpticalNode` instance, or a data type which can be converted
        to an optical node.

    direction : str
        Plane to detect in - 'x' for tangential, 'y' for sagittal. Defaults to 'x'.
    """

    def __init__(
        self, name, *args, from_node=None, to_node=None, via_node=None, direction="x"
    ):
        Detector.__init__(
            self, name, dtype=np.float64, unit="degrees", label="Gouy phase"
        )

        nodes = from_node, to_node, via_node
        if args and any(node is not None for node in nodes):
            raise ValueError(
                f"Error in Gouy detector {self.name}:\n"
                "    Cannot specify both a sequence of spaces and a node path."
            )

        if not args and all(node is None for node in nodes):
            raise ValueError(
                f"Error in Gouy detector {self.name}:\n"
                "    Must specify EITHER a sequence of spaces or a node path."
            )

        self.direction = direction
        self.__spaces = list(args)
        self.__from_node = from_node
        self.__to_node = to_node
        self.__via_node = via_node

    @property
    def needs_fields(self):
        return False

    @property
    def needs_trace(self):
        return True

    @property
    def spaces(self):
        """The spaces that this detector will calculated the accumulated Gouy phase
        over."""
        return self.__spaces.copy()

    def _lookup_spaces(self):
        from finesse.components import Space

        if not self.has_model:
            raise RuntimeError(f"Bug detected! No model associated with {self.name}")

        for i, item in enumerate(
            self.__spaces
        ):  # sequence of spaces / space names given
            if isinstance(item, Space):
                continue

            if not isinstance(item, str):
                raise TypeError(
                    f"Error in Gouy detector {self.name}:\n"
                    "    Unexpected type in iterable of spaces. Expected a "
                    f"Space or str, but got type: {type(item)}"
                )

            space = self._model.elements.get(item, None)
            if space is None:
                raise LookupError(
                    f"Error in Gouy detector {self.name}:\n"
                    f"    No space of name {item} in model."
                )
            if not isinstance(space, Space):
                raise TypeError(
                    f"Error in Gouy detector {self.name}:\n"
                    f"    Element of name {item} is not a Space."
                )

            self.__spaces[i] = space

        # node path source, target given
        if self.__from_node is not None and self.__to_node is not None:
            path = self._model.path(self.__from_node, self.__to_node, self.__via_node)
            self.__spaces = list(path.spaces)

        if self.__spaces:
            LOGGER.info(
                "Gouy detector %s using spaces: %s", self.name, self.__spaces,
            )
        else:
            LOGGER.warning(
                "Gouy detector %s not using any spaces! This will always return "
                "values of zero in a simulation.",
                self.name,
            )

    def _get_workspace(self, sim):
        ws = GouyDetectorWorkspace(self, sim)
        return ws



import numpy as np
from numpy.testing import assert_allclose
import pytest
from finesse import Model
from finesse.analysis import noxaxis
from finesse.detectors import camera
from finesse.gaussian import HGMode


@pytest.fixture
def fp_cavity_model():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    s s0 L0.p1 ITM.p1

    m ITM R=0.9 T=0.1 Rc=-2
    s sc ITM.p2 ETM.p1 L=1
    m ETM R=0.9 T=0.1 Rc=2

    cav FP ITM.p2
    """
    )
    IFO.ITM.Rcy = IFO.ITM.Rcx.ref
    IFO.ETM.Rcy = IFO.ETM.Rcx.ref

    return IFO


@pytest.mark.parametrize("ltype", ("fpx", "ccdpx"))
@pytest.mark.parametrize("x", (0, 0.3, -1.2, np.pi))
@pytest.mark.parametrize("y", (0, 0.7, -2.4, -np.pi))
@pytest.mark.parametrize("w0_scaled", (True, False))
def test_pixel_tem00_circ(fp_cavity_model: Model, ltype, x, y, w0_scaled):
    """Test that pixel camera type outputs for a circulating field which should be a
    pure HG00 mode corresponds to product of u_00 and a_00.

    Tested for both FieldPixel and CCDPixel --- i.e. amp + phase and intensity.
    """
    IFO = fp_cavity_model
    IFO.select_modes(maxtem=2)
    IFO.parse(f"{ltype} px ITM.p2.o x={x} y={y} w0_scaled={int(w0_scaled)}")

    IFO.parse("ad ad00 ITM.p2.o f=0 n=0 m=0")

    out = noxaxis(IFO)

    q_cav = IFO.FP.qx
    HG00 = HGMode(q_cav)

    x = IFO.px.x
    y = IFO.px.y
    if w0_scaled:
        x *= q_cav.w0
        y *= q_cav.w0

    u_nm = HG00.un(x) * HG00.um(y)
    if ltype == "fpx":
        expect = u_nm * out["ad00"]
    else:
        expect = np.abs(u_nm) ** 2 * np.abs(out["ad00"]) ** 2

    assert_allclose(out["px"], expect)


@pytest.mark.parametrize("ltype", (camera.FieldScanLine, camera.CCDScanLine))
@pytest.mark.parametrize("d", ("x", "y"))
@pytest.mark.parametrize("other_ax_offset", (0, 0.5, -1.4))
@pytest.mark.parametrize("w0_scaled", (True, False))
def test_scanline_tem00_circ(
    fp_cavity_model: Model, ltype, d, other_ax_offset, w0_scaled
):
    """Test that scan-line camera type outputs for a circulating field which should be a
    pure HG00 mode corresponds to product of u_00 and a_00.

    Tested for both FieldLine and CCDScanLine --- i.e. amp + phase and intensity.
    """
    IFO = fp_cavity_model
    IFO.select_modes(maxtem=2)
    if d == "x":
        IFO.add(
            ltype(
                "line",
                IFO.ITM.p2.o,
                xlim=3,
                y=other_ax_offset,
                npts=100,
                w0_scaled=w0_scaled,
            )
        )
    else:
        IFO.add(
            ltype(
                "line",
                IFO.ITM.p2.o,
                x=other_ax_offset,
                ylim=3,
                npts=100,
                w0_scaled=w0_scaled,
            )
        )

    IFO.parse("ad ad00 ITM.p2.o f=0 n=0 m=0")

    out = noxaxis(IFO)

    q_cav = IFO.FP.qx
    HG00 = HGMode(q_cav)
    if d == "x":
        x = IFO.line.xdata
        y = other_ax_offset
    else:
        x = other_ax_offset
        y = IFO.line.ydata

    if w0_scaled:
        x *= q_cav.w0
        y *= q_cav.w0

    u_nm = HG00.un(x) * HG00.um(y)
    if ltype is camera.FieldScanLine:
        expect = u_nm * out["ad00"]
    else:
        expect = np.abs(u_nm) ** 2 * np.abs(out["ad00"]) ** 2

    assert_allclose(out["line"], expect)


@pytest.mark.parametrize("ltype", ("fcam", "ccd"))
@pytest.mark.parametrize("w0_scaled", (True, False))
def test_image_tem00_circ(fp_cavity_model: Model, ltype, w0_scaled):
    """Test that image camera outputs for pure circulating TEM00 mode.

    correspond to u_00 * a_00 (for fcam) and |u_00|^2 |a_00|^2 (for ccd).
    """
    IFO = fp_cavity_model
    IFO.select_modes(maxtem=2)
    IFO.parse(
        f"{ltype} image ITM.p2.o xlim=5 ylim=5 npts=60 w0_scaled={int(w0_scaled)}"
    )

    IFO.parse("ad ad00 ITM.p2.o f=0 n=0 m=0")

    out = noxaxis(IFO)

    q_cav = IFO.FP.qx
    HG00 = HGMode(q_cav)

    x = IFO.image.xdata
    y = IFO.image.ydata
    if w0_scaled:
        x *= q_cav.w0
        y *= q_cav.w0

    u_nm = HG00.unm(x, y)
    if ltype == "fcam":
        expect = u_nm * out["ad00"]
    else:
        expect = u_nm * np.conjugate(u_nm) * np.abs(out["ad00"]) ** 2

    assert_allclose(out["image"], expect)



"""
Optical components performing directional suppression of beams.
"""

import logging

import numpy as np

from finesse.components.modal.isolator import IsolatorWorkspace
from finesse.components.general import Connector, InteractionType
from finesse.components.node import NodeDirection, NodeType
from finesse.parameter import float_parameter
from finesse.utilities import refractive_index

LOGGER = logging.getLogger(__name__)


@float_parameter("S", "Power suppression")
class Isolator(Connector):
    """Represents an isolator optical component with a suppression factor.

    Parameters
    ----------
    name : str
        Name of newly created isolator.

    S : float
        Power suppression in dB.
    """

    def __init__(self, name, S=0.0):
        super().__init__(name)

        self.S = S

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        # optic to optic couplings
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

    def _get_workspace(self, sim):
        _, is_changing = self._eval_parameters()
        refill = sim.is_component_in_mismatch_couplings(self) or len(is_changing)

        ws = IsolatorWorkspace(self, sim, refill)
        # This assumes that nr1/nr2 cannot change during a simulation
        ws.nr1 = refractive_index(self.p1)
        ws.nr2 = refractive_index(self.p2)

        ws.add_fill_function(self._fill_matrix, refill)

        if sim.is_modal:
            ws.set_knm_info(
                "P1i_P2o", nr_from=ws.nr1, nr_to=ws.nr2, is_transmission=True
            )
            ws.set_knm_info(
                "P2i_P1o", nr_from=ws.nr2, nr_to=ws.nr1, is_transmission=True
            )

        return ws

    def _fill_matrix(self, ws):
        values, _ = self._eval_parameters()
        suppression = 10 ** (-values["S"] / 20)

        for freq in ws.sim.frequencies:
            with ws.sim.component_edge_fill3(
                ws.owner_id, ws.connections.P1i_P2o_idx, freq.index, freq.index
            ) as mat:
                mat[:] = ws.K12.data

            with ws.sim.component_edge_fill3(
                ws.owner_id, ws.connections.P2i_P1o_idx, freq.index, freq.index,
            ) as mat:
                np.multiply(suppression, ws.K21.data, out=mat[:])



"""Optical cavities with associated properties."""

from copy import deepcopy
import cmath
import logging
import math

import networkx as nx
import numpy as np

from finesse.config import config_instance
import finesse.components as components
import finesse.tracing.ctracer as ctracer
from finesse.components.node import OpticalNode, Port
from finesse.cymath.math import sgn
from finesse.constants import values as constants
from finesse.parameter import info_parameter
from finesse.gaussian import BeamParam

from .trace_dependency import TraceDependency


LOGGER = logging.getLogger(__name__)

# TODO (sjr) Add Cavity.gain property
#            - optical gain eqn? simple for 2 mirror cav, but need general eqn.


@info_parameter("FSR", "FSR", units="Hz")
@info_parameter("loss", "Loss")
@info_parameter("finesse", "Finesse")
@info_parameter("FWHM", "FWHM", units="Hz")
@info_parameter("storage_time", "Storage time", units="s")
@info_parameter("pole", "Pole", units="Hz")
@info_parameter("round_trip_length", "Round trip length", units="m")
@info_parameter("w0", "Waist size", units="m")
@info_parameter("waistpos", "Waist position", units="m")
@info_parameter("m", "Stability (m-factor)")
@info_parameter("g", "Stability (g-factor)")
@info_parameter("gouy", "Round trip gouy phase", units="°")
@info_parameter("mode_separation", "Mode separation", units="Hz")
@info_parameter("S", "Resolution")
@info_parameter("is_stable", "Stable")
@info_parameter("is_critical", "Critically stable")
class Cavity(TraceDependency):
    """Represents a cavity in an interferometer configuration.

    This class stores the shortest path between the start node and end node of the cavity and
    holds symbolic expressions for each physical attribute of the cavity. Numeric values
    corresponding to these attributes are obtained through the relevant properties.

    Adding a Cavity to a :class:`.Model` results in the beam parameters of all nodes in the cavity
    path being set according to the cavity eigenmode (:attr:`Cavity.q`) when a beam trace is
    performed (e.g. at the start of a modal based simulation). The mode of the cavity is then also
    used as a trace starting point when setting beam parameters at nodes outside of the cavity - see
    :ref:`tracing_manual` for details on the beam tracing algorithm.

    Parameters
    ----------
    name : str
        Name of newly created cavity.

    source : :class:`.OpticalNode` or :class:`.Port`
        Node / Port that the cavity path starts from. If no `via` node is specified, then the cavity
        path will be given by the shortest path from source back to the component that owns
        source.

        If a port is given then the *output* optical node of that port will be used as the source.

    via : :class:`.OpticalNode`, optional
        Node that the cavity path must traverse via; defaults to `None`.

        Note that, unlike `source`, this cannot be a :class:`.Port` object as this would be ambiguous
        for beamsplitter type components - i.e. determination of which node to use cannot be
        assumed automatically.

    priority : number, optional; default: 0
        Priority value for beam tracing. Beam tracing dependencies are sorted in descending order
        of priority - i.e. higher priority value dependencies will be traced first. Any dependency
        with a priority value of zero will be traced, after non-zero priority dependencies, in alphabetic
        order of the dependency names.
    """

    def __init__(self, name, source, via=None, priority=0):
        super().__init__(name, priority)

        # If the source is a port then use the output node
        # of this port as the source for the cavity path
        if isinstance(source, Port):
            source = source.o

        if not isinstance(source, OpticalNode):
            raise TypeError("Expected source to be an OpticalNode.")
        else:
            if source.is_input:
                raise ValueError("Source must be an output node.")

            if not isinstance(source.component, components.Surface):
                msg = (
                    "Expected owner of source node to be a Surface, "
                    f"but got a {type(self.source.component)}"
                )
                raise TypeError(msg)

        if via is not None and not isinstance(via, OpticalNode):
            raise TypeError("Expected via to be an OpticalNode.")

        self.__source = source
        self.__via = via
        self.__path = None

    def __deepcopy__(self, memo):
        # TraceTree instances are non-copyable by design so
        # temporarily set __tree field to None to avoid copy
        tmp = self.__tree
        self.__tree = None
        new = super().__deepcopy__(memo)

        memo[id(self)] = new

        new.__dict__.update(deepcopy(self.__dict__, memo))
        new_model = memo.get(id(self._model))
        new._reset_model(new_model)

        def update_later():
            new.initialise()

        new_model.after_deepcopy.append(update_later)

        # Re-build the tree on self due to above - not an
        # expensive operation so no big deal for now
        self.__tree = tmp
        return new

    def draw(self):
        """A string representation of the cavity route.

        Returns
        -------
        s : str
            The node path of the cavity as a string.
        """
        return self.__tree.draw()

    @property
    def path(self):
        """The :class:`.OpticalPath` instance of the cavity.

        :getter: Returns the path of the cavity (read-only).

        See Also
        --------
        finesse.model.Model.path : Retrieves an ordered container of the path trace between two
                                   specified nodes.
        """
        return self.__path

    @property
    def source(self):
        """Starting node of the cavity.

        :getter: Returns the cavity starting node (read-only).
        """
        return self.__source

    @property
    def via(self):
        """Via node of the cavity.

        :getter: Returns the cavity via node (read-only).
        """
        return self.__via

    @property
    def is_fabry_perot(self):
        """Flag indicating whether the cavity is a Fabry-Perot cavity.

        :getter: Returns true if the cavity is a Fabry-Perot, false otherwise (read-only).
        """
        if self.path is None:
            return False

        unique_spaces = set(self.path.spaces)
        # cavity is Fabry-Perot if there's only one space in the path
        return len(unique_spaces) == 1

    ### Non-geometric properties ###

    @property
    def FSR(self):
        r"""The free-spectral-range (FSR) of the cavity.

        This quantity is defined as,

        .. math::
            \mathrm{FSR} = \frac{c}{2L},

        where :math:`c` is the speed of light and :math:`L` is the
        length of the cavity.

        :getter: Returns the cavity free-spectral-range (read-only).
        """
        return self._FSR.eval()

    @property
    def loss(self):
        r"""The round-trip loss of the cavity as a fraction of the incoming power.

        This quantity is computed via,

        .. math::
            L = 1 - \prod_{\mathrm{i}}^{N_{\mathrm{refl}}} R_{\mathrm{i}} \times
                \prod_{\mathrm{i}}^{N_{\mathrm{trns}}} T_{\mathrm{i}},

        i.e. one minus the product of all reflections multiplied with the product of all
        transmissions for a round-trip of the cavity.

        :getter: Returns the fractional round-trip cavity loss (read-only).
        """
        return self._loss.eval()

    @property
    def finesse(self):
        r"""The finesse of the cavity.

        This quantity is defined as,

        .. math::
            \mathcal{F} = \frac{\pi}{2\arcsin{\left(
                \frac{1 - \widetilde{l}}{2\sqrt{\widetilde{l}}}
            \right)
            }},

        where :math:`\widetilde{l} = \sqrt{1 - l}` and :math:`l`
        is the cavity loss.

        :getter: Returns the cavity finesse (read-only).
        """
        return self._finesse.eval()

    @property
    def FWHM(self):
        r"""The cavity full-width-half-maximum (FWHM).

        This quantity is defined as,

        .. math::
            \mathrm{FWHM} = \frac{\mathrm{FSR}}{\mathcal{F}},

        where :math:`\mathcal{F}` is the cavity finesse.

        :getter: Returns the FWHM of the cavity (read-only).

        See Also
        --------
        Cavity.FSR : Free-spectral-range of a cavity.
        Cavity.finesse : Finesse of a cavity.
        """
        return self._FWHM.eval()

    @property
    def storage_time(self):
        r"""The cavity storage time (:math:`\tau`).

        This quantity is defined as,

        .. math::
            \tau = \frac{1}{\pi\mathrm{FWHM}},

        where :math:`\mathrm{FWHM}` is the full-width at half-maximum
        of the cavity resonance.

        :getter: Returns the storage time of the cavity (read-only).

        See Also
        --------
        Cavity.FWHM : Full-width at half-maximum (FWHM) of a cavity.
        """
        return self._tau.eval()

    @property
    def pole(self):
        r"""The pole-frequency of the cavity.

        This quantity is defined as,

        .. math::
            f_{\mathrm{pole}} = \frac{\mathrm{FWHM}}{2},

        where :math:`\mathrm{FWHM}` is the full-width at half-maximum
        of the cavity resonance.

        :getter: Returns the cavity pole-frequency (read-only).

        See Also
        --------
        Cavity.FWHM : Full-width at half-maximum (FWHM) of a cavity.
        """
        return self._pole.eval()

    ### Geometric properties ###

    @property
    def round_trip_length(self):
        """The round-trip length of the cavity (in metres).

        :getter: Returns the length of a single round-trip of the cavity (read-only).
        """
        return self._length.eval()

    @property
    def ABCD(self):
        """The round-trip ABCD matrix of the cavity in both plannes.

        :getter: Returns a :class:`numpy.ndarray` with shape ``(2, 2, 2)`` of the cavity
                 round-trip matrices in the tangential and sagittal planes, respectively. (read-only).
        """
        ctracer.update_rt_abcd(self.__tree, self._ABCDx, self._ABCDy)
        Ms = np.zeros((2, 2, 2))
        Ms[0][:] = self._ABCDx[:]
        Ms[1][:] = self._ABCDy[:]
        return Ms

    def __get_ABCD(self, direction):
        return getattr(self, f"ABCD{direction}")

    @property
    def ABCDx(self):
        """The tangential round-trip ABCD matrix of the cavity.

        :getter: Returns the cavity round-trip matrix in the
                 tangential plane (read-only).
        """
        ctracer.update_single_rt_abcd(self.__tree, self._ABCDx, "x")
        return self._ABCDx

    @property
    def ABCDy(self):
        """The sagittal round-trip ABCD matrix of the cavity.

        :getter: Returns the cavity round-trip matrix in the
                 sagittal plane (read-only).
        """
        ctracer.update_single_rt_abcd(self.__tree, self._ABCDy, "y")
        return self._ABCDy

    @property
    def q(self):
        r"""The eigenmode of the cavity in both planes.

        For a single plane, the cavity eigenmode :math:`q_{\mathrm{cav}}` is computed by solving,

        .. math::
            C q_{\mathrm{cav}}^2+(D-A)q_{\mathrm{cav}} - B = 0,

        where :math:`A`, :math:`B`, :math:`C` and :math:`D` are the elements of the
        round-trip ABCD matrix of the cavity for this plane.

        :getter: Returns a :class:`numpy.ndarray` of the cavity eigenmodes in the tangential and
                 sagittal planes, respectively, where both values are :class:`.BeamParam`
                 instances. (read-only).
        """
        return np.array([self.qx, self.qy])

    def __get_lambda0(self):
        if self.has_model:
            lambda0 = self._model.lambda0
        else:
            lambda0 = config_instance()["constants"].getfloat("lambda0")

        return lambda0

    def __compute_eigenmode(self, direction):
        ABCD = self.__get_ABCD(direction)

        C = ABCD[1][0]
        if C == 0.0:  # confocal cavity - g = 0 (critical)
            return None
        half_inv_C = 0.5 / C

        D_minus_A = ABCD[1][1] - ABCD[0][0]
        minus_B = -1 * ABCD[0][1]

        sqrt_term = cmath.sqrt(D_minus_A * D_minus_A - 4 * C * minus_B)
        lower = (-D_minus_A - sqrt_term) * half_inv_C
        upper = (-D_minus_A + sqrt_term) * half_inv_C

        if lower.imag > 0:
            q = lower
        elif upper.imag > 0:
            q = upper
        else:
            return None

        return q

    @property
    def qx(self):
        """The eigenmode of the cavity in the tangential plane.

        :getter: Returns the cavity's tangential plane eigenmode (read-only).

        See Also
        --------
        Cavity.q
        """
        q = self.__compute_eigenmode("x")
        if q is None:
            return None

        return BeamParam(q=q, wavelength=self.__get_lambda0())

    @property
    def qy(self):
        """The eigenmode of the cavity in the sagittal plane.

        :getter: Returns the cavity's sagittal plane eigenmode (read-only).

        See Also
        --------
        Cavity.q
        """
        q = self.__compute_eigenmode("y")
        if q is None:
            return None

        return BeamParam(q=q, wavelength=self.__get_lambda0())

    @property
    def w0(self):
        """The waist size of the cavity in both planes.

        :getter: Returns a :class:`numpy.ndarray` of the cavity waist size in the tangential
                 and sagittal planes, respectively. (read-only).
        """
        return np.array([self.w0x, self.w0y])

    @property
    def w0x(self):
        """The waist size of the cavity in the tangential plane.

        Equivalent to ``cavity.qx.w0``.

        :getter: Returns the cavity waist size in the tangential plane (read-only).
        """
        if not self.is_stable_x:
            return np.nan

        return self.qx.w0

    @property
    def w0y(self):
        """The waist size of the cavity in the sagittal plane.

        Equivalent to ``cavity.qy.w0``.

        :getter: Returns the cavity waist size in the sagittal plane (read-only).
        """
        if not self.is_stable_y:
            return np.nan

        return self.qy.w0

    @property
    def waistpos(self):
        """The position of the cavity waist in both planes.

        This distance to the waist is measured using the position of :attr:`Cavity.source`
        node as the origin.

        :getter: Returns a :class:`numpy.ndarray` of the cavity waist position in the tangential
                 and sagittal planes, respectively. (read-only).
        """
        return np.array([self.waistpos_x, self.waistpos_y])

    @property
    def waistpos_x(self):
        """The waist position of the cavity in the tangential plane.

        Equivalent to ``cavity.qx.z``.

        :getter: Returns the cavity waist position in the tangential plane (read-only).
        """
        if not self.is_stable_x:
            return np.nan

        return self.qx.z

    @property
    def waistpos_y(self):
        """The waist position of the cavity in the sagittal plane.

        Equivalent to ``cavity.qy.z``.

        :getter: Returns the cavity waist position in the sagittal plane (read-only).
        """
        if not self.is_stable_y:
            return np.nan

        return self.qy.z

    @property
    def m(self):
        r"""The stability of the cavity, in both planes, given by the :math:`m`-factor:

        .. math::
            m = \frac{A + D}{2},

        where :math:`A` and :math:`D` are the relevant entries of the
        cavity round-trip ABCD matrix. The cavity is stable if the
        following condition is satisfied:

        .. math::
            -1 \leq m \leq 1.

        :getter: Returns a :class:`numpy.ndarray` of the cavity stability in the tangential and
                 sagittal planes, respectively. (read-only).
        """
        return np.array([self.mx, self.my])

    def __compute_m(self, direction):
        ABCD = self.__get_ABCD(direction)

        A = ABCD[0][0]
        D = ABCD[1][1]
        return 0.5 * (A + D)

    @property
    def mx(self):
        """The stability, m, of the cavity in the tangential plane.

        :getter: Returns the tangential plane m-factor (read-only).

        See Also
        --------
        Cavity.m
        Cavity.gx
        """
        return self.__compute_m("x")

    @property
    def my(self):
        """The stability, m, of the cavity in the sagittal plane.

        :getter: Returns the sagittal plane m-factor (read-only).

        See Also
        --------
        Cavity.m
        Cavity.gy
        """
        return self.__compute_m("y")

    @property
    def g(self):
        r"""The stability of the cavity, in both planes, given by the :math:`g`-factor:

        .. math::
            g = \frac{A + D + 2}{4},

        where :math:`A` and :math:`D` are the relevant entries of the
        cavity round-trip ABCD matrix. The cavity is stable if the
        following condition is satisfied:

        .. math::
            0 \leq g \leq 1.

        :getter: Returns a :class:`numpy.ndarray` of the cavity stability in the tangential and
                 sagittal planes, respectively. (read-only).
        """
        return np.array([self.gx, self.gy])

    def __compute_g(self, direction):
        m = self.__compute_m(direction)
        return 0.5 * (1 + m)

    @property
    def gx(self):
        """The stability, g, of the cavity in the tangential plane.

        :getter: Returns the tangential plane g-factor (read-only).

        See Also
        --------
        Cavity.g
        Cavity.mx
        """
        return self.__compute_g("x")

    @property
    def gy(self):
        """The stability, g, of the cavity in the sagittal plane.

        :getter: Returns the sagittal plane g-factor (read-only).

        See Also
        --------
        Cavity.g
        Cavity.my
        """
        return self.__compute_g("y")

    @property
    def gouy(self):
        r"""The accumulated round-trip Gouy phase of the cavity in both planes (in degrees).

        This is given by,

        .. math::
            \psi_{\mathrm{rt}} = 2\,\arccos{\left( \mathrm{sgn}(B) \sqrt{g} \right)},

        where :math:`B` is the corresponding element of the round-trip
        ABCD matrix and :math:`g` is the cavity stability parameter
        returned by :attr:`Cavity.g`.

        :getter: Returns a :class:`numpy.ndarray` of the accumulated round-trip Gouy phase in the
                 tangential and sagittal planes, respectively. (read-only).
        """
        return np.array([self.gouy_x, self.gouy_y])

    def __compute_round_trip_gouy(self, direction, deg=True):
        ABCD = self.__get_ABCD(direction)

        B = ABCD[0][1]
        g = getattr(self, f"g{direction}")

        psi = 2 * math.acos(sgn(B) * math.sqrt(g))

        if deg:
            return math.degrees(psi)
        return psi

    @property
    def gouy_x(self):
        """The round-trip Gouy phase in the tangential plane (in degrees).

        If the cavity is not stable, then ``np.nan`` is returned.

        :getter: Returns the tangential plane round-trip Gouy phase (read-only).

        See Also
        --------
        Cavity.gouy
        """
        if not self.is_stable_x:
            return np.nan

        return self.__compute_round_trip_gouy("x")

    @property
    def gouy_y(self):
        """The round-trip Gouy phase in the sagittal plane (in degrees).

        If the cavity is not stable, then ``np.nan`` is returned.

        :getter: Returns the sagittal plane round-trip Gouy phase (read-only).

        See Also
        --------
        Cavity.gouy
        """
        if not self.is_stable_y:
            return np.nan

        return self.__compute_round_trip_gouy("y")

    @property
    def mode_separation(self):
        r"""The mode separation frequency of the cavity in both planes.

        This is defined by,

        .. math::
            \delta f =
                \begin{cases}
                    \frac{\psi_{\mathrm{rt}}}{2\pi}
                    \Delta f, & \text{if } \psi_{\mathrm{rt}} \leq \pi\\
                    (1 - \frac{\psi_{\mathrm{rt}}}{2\pi}) \Delta f, & \text{if } \psi_{\mathrm{rt}}
                    > \pi,
                \end{cases}

        where :math:`\psi_{\mathrm{rt}}` is the accumulated round-trip Gouy phase
        and :math:`\Delta f` is the FSR of the cavity.

        :getter: Returns a :class:`numpy.ndarray` of the mode separation frequency in the tangential
                 and sagittal planes, respectively. (read-only).
        """
        return np.array([self.mode_separation_x, self.mode_separation_y])

    def __compute_mode_separation(self, direction):
        gouy = self.__compute_round_trip_gouy(direction, deg=False)

        fsr = self.FSR
        df = 0.5 * fsr * gouy / np.pi
        if gouy > np.pi:
            df = fsr - df

        return df

    @property
    def mode_separation_x(self):
        """The mode separation frequency in the tangential plane.

        If the cavity is not stable, then ``np.nan`` is returned.

        :getter: Returns the tangential plane mode separation frequency (read-only).

        See Also
        --------
        Cavity.mode_separation
        """
        if not self.is_stable_x:
            return np.nan

        return self.__compute_mode_separation("x")

    @property
    def mode_separation_y(self):
        """The mode separation frequency in the sagittal plane.

        If the cavity is not stable, then ``np.nan`` is returned.

        :getter: Returns the sagittal plane mode separation frequency (read-only).

        See Also
        --------
        Cavity.mode_separation
        """
        if not self.is_stable_y:
            return np.nan

        return self.__compute_mode_separation("y")

    @property
    def S(self):
        r"""The resolution of the cavity in both planes.

        Cavity resolution, :math:`S`, is defined by,

        .. math::
            S =
                \begin{cases}
                    \frac{\psi_{\mathrm{rt}}}{2\pi}
                    \mathcal{F}, & \text{if } \psi_{\mathrm{rt}} \leq \pi\\
                    (1 - \frac{\psi_{\mathrm{rt}}}{2\pi})
                        \mathcal{F}, & \text{if } \psi_{\mathrm{rt}} > \pi,
                \end{cases}

        where :math:`\psi_{\mathrm{rt}}` is the round-trip Gouy phase and :math:`\mathcal{F}` is the
        cavity finesse.

        :getter: Returns a :class:`numpy.ndarray` of the cavity resolution in the tangential and
                 sagittal planes, respectively. (read-only).
        """
        return np.array([self.Sx, self.Sy])

    def __compute_resolution(self, direction):
        gouy = self.__compute_round_trip_gouy(direction, deg=False)

        f = self.finesse
        s = 0.5 * f * gouy / np.pi
        if gouy > np.pi:
            s = f - s

        return s

    @property
    def Sx(self):
        """The resolution of cavity in the tangential plane.

        If the cavity is not stable, then ``np.nan`` is returned.

        :getter: Returns the tangential plane resolution (read-only).

        See Also
        --------
        Cavity.S
        """
        if not self.is_stable_x:
            return np.nan

        return self.__compute_resolution("x")

    @property
    def Sy(self):
        """The resolution of cavity in the sagittal plane.

        If the cavity is not stable, then ``np.nan`` is returned.

        :getter: Returns the sagittal plane resolution (read-only).

        See Also
        --------
        Cavity.S
        """
        if not self.is_stable_y:
            return np.nan

        return self.__compute_resolution("y")

    ### Stability flags ###

    @property
    def is_stable(self):
        r"""Flag indicating whether the cavity is stable.

        This only returns `True` if *both* planes of the cavity eigenmode
        are stable.

        :getter: Returns `True` if :math:`0 \leq g \leq 1`,
                 `False` otherwise (for both tangential, sagittal planes).

        See Also
        --------
        Cavity.g
        """
        return self.is_stable_x and self.is_stable_y

    @property
    def is_stable_x(self):
        r"""Flag indicating whether cavity is stable in the tangential plane.

        :getter: Returns `True` if :math:`0 \leq g_x \leq 1`, `False` otherwise.

        See Also
        --------
        Cavity.is_stable
        Cavity.gx
        """
        return 0 < self.gx < 1

    @property
    def is_stable_y(self):
        r"""Flag indicating whether cavity is stable in the sagittal plane.

        :getter: Returns `True` if :math:`0 \leq g_y \leq 1`, `False` otherwise.

        See Also
        --------
        Cavity.is_stable
        Cavity.gy
        """
        return 0 < self.gy < 1

    @property
    def is_critical(self):
        r"""Flag indicating whether the cavity is critically stable.

        This only returns `True` if *both* planes of the cavity eigenmode
        are critically stable.

        :getter: Returns `True` if :math:`g = 0` or :math:`g = 1`,
                 `False` otherwise (for both tangential, sagittal planes).

        See Also
        --------
        Cavity.g
        """
        return self.is_critical_x and self.is_critical_y

    @property
    def is_critical_x(self):
        r"""Flag indicating whether the cavity is critically stable in the tangential plane.

        :getter: Returns `True` if :math:`g_x = 0` or :math:`g_x = 1`,
                 `False` otherwise.

        See Also
        --------
        Cavity.is_critical
        Cavity.gx
        """
        gx = self.gx
        return gx == 0 or gx == 1

    @property
    def is_critical_y(self):
        r"""Flag indicating whether the cavity is critically stable in the sagittal plane.

        :getter: Returns `True` if :math:`g_x = 0` or :math:`g_x = 1`,
                 `False` otherwise.

        See Also
        --------
        Cavity.is_critical
        Cavity.gy
        """
        gy = self.gy
        return gy == 0 or gy == 1

    ### Methods for setting up the symbolic equations for each property ###

    def __symbolise_round_trip_length(self):
        self._length = 0.0
        for comp in self.__path.spaces:
            self._length += comp.L.ref

    def __symbolise_FSR(self):
        self._FSR = constants.C_LIGHT / self._length

    def __symbolise_loss(self):
        from finesse.components.general import InteractionType

        power = 1.0
        t = self.__tree
        while t.left is not None:
            if t.node.is_input:
                comp = t.node.component
            else:
                comp = t.node.space

            if isinstance(comp, components.Surface):
                if (
                    comp.interaction_type(t.node, t.left.node)
                    == InteractionType.REFLECTION
                ):
                    power *= comp.R.ref
                else:
                    power *= comp.T.ref

            t = t.left

        # Need to do final reflection from root component
        power *= self.__tree.node.component.R.ref
        self._loss = 1.0 - power

    def __symbolise_finesse(self):
        _loss = np.sqrt(1.0 - self._loss)
        self._finesse = 0.5 * np.pi / np.arcsin(0.5 * (1.0 - _loss) / np.sqrt(_loss))

    def __symbolise_FWHM(self):
        self._FWHM = self._FSR / self._finesse

    def __symbolise_storage_time(self):
        self._tau = 1 / (np.pi * self._FWHM)

    def __symbolise_pole(self):
        self._pole = 0.5 * self._FWHM

    def __initialise_ABCD(self):
        self.__tree = ctracer.TraceTree.from_cavity(self)

        self._ABCDx = np.eye(2)
        self._ABCDy = np.eye(2)

        ctracer.update_rt_abcd(self.__tree, self._ABCDx, self._ABCDy)

    def _get_workspace(self, sim):
        from finesse.components.modal.cavity import CavityWorkspace

        return CavityWorkspace(self, sim)

    def initialise(self):
        """Initialises the symbolic equations of the cavity and calculates the cavity
        path from the associated model."""

        # determine the target node of the cavity path
        if isinstance(self.source.component, components.Mirror):
            target = self.source.opposite
        else:
            bs = self.source.component
            if self.source.port is bs.p1:
                target = bs.p2.i
            elif self.source.port is bs.p2:
                target = bs.p1.i
            elif self.source.port is bs.p3:
                target = bs.p4.i
            else:
                target = bs.p3.i

        self.__path = self._model.path(self.source, target, self.via)

        self.__initialise_ABCD()
        # Set up all the symbolic equations for each property
        self.__symbolise_round_trip_length()
        self.__symbolise_loss()
        self.__symbolise_finesse()
        self.__symbolise_FSR()
        self.__symbolise_FWHM()
        self.__symbolise_storage_time()
        self.__symbolise_pole()

        if self.is_fabry_perot:
            comps = list(
                filter(
                    lambda x: isinstance(x, components.Surface),
                    self.path.components_only,
                )
            )
            M1, M2 = comps[0], comps[-1]

            R1x, R1y = M1.Rcx.value, M1.Rcy.value
            R2x, R2y = M2.Rcx.value, M1.Rcy.value
            if sgn(R1x) == sgn(R2x) and sgn(R1y) == sgn(R2y) and not self.is_stable:
                LOGGER.warning(
                    "The signs of the radii of curvature of mirrors "
                    "%s and %s in the unstable Fabry-Perot cavity %s are equal.",
                    M1.name,
                    M2.name,
                    self.name,
                )

    def any_changing_params(self, geometric=False):
        """Determines whether any parameter of any component inside the cavity is
        changing.

        If the optional argument `geometric` is True, then this will only
        check that the following parameters are changing:

        - radii of curvature of surfaces,
        - lengths of spaces,
        - refractive indices of spaces,
        - focal lengths of lenses,
        - angles of incidence of beam splitters.

        Parameters
        ----------
        geometric : bool
            If true then only checks parameters which affect ABCD matrices.

        Returns
        -------
        flag : bool
            True if any parameter is changing (subject to the condition outlined
            above), False otherwise.
        """
        geometric_params = ["Rcx", "Rcy", "L", "f", "nr", "alpha"]
        for comp in self.path.components_only:
            for param in comp.parameters:
                if param.is_changing:
                    if geometric and param.name not in geometric_params:
                        continue

                    return True

        return False

    @property
    def is_changing(self):
        """Flag indicating whether any geometric parameter inside the cavity is
        changing.

        A geometric parameter is defined as one of:

        - radii of curvature of surfaces,
        - focal lengths of lenses,
        - angles of incidence of beam splitters,
        - lengths of spaces,
        - refractive indices of spaces,

        i.e. any parameter which can affect the ABCD matrix values.
        """
        return self.any_changing_params(geometric=True)

    def get_exit_nodes(self):
        """Obtains a dictionary of `source: target` mappings where source -> target and
        target is an exit node of the cavity.

        An exit node is defined to be a node that is not internal
        to the cavity, rather it is obtained on propagation from
        an internal node to outside the cavity.

        Returns
        -------
        exit_nodes : dict
            A dictionary of `source: target` mappings.
        """
        source_succ = nx.dfs_successors(
            self._model.optical_network, self.source.full_name
        )

        node_from_name = lambda n: self._model.network.nodes[n]["weakref"]()
        cav_nodes = self.path.nodes_only
        exit_nodes = {}
        for source_name, target_names in source_succ.items():
            source = node_from_name(source_name)
            if source in cav_nodes:
                for target_name in target_names:
                    target = node_from_name(target_name)
                    if target not in cav_nodes:
                        exit_nodes[source] = target

        return exit_nodes

    def trace_beam(self):
        """Traces the cavity eigenmode through the cavity path.

        Returns
        -------
        out : :class:`.BeamTraceSolution`
            An object representing the results of the tracing routine.
        """
        from ..solutions import BeamTraceSolution

        trace = {}
        wavelength = self.__get_lambda0()
        self.__tree.trace_beam(trace, wavelength, symmetric=True)

        return BeamTraceSolution(f"{self.name}_trace", trace, [self.__tree])

    def plot(self, direction=None, *args, **kwargs):
        """Plots the beam representing the cavity eigenmode over the path of the cavity.

        See :meth:`.PropagationSolution.plot` if specifying `direction` or
        :meth:`.AstigmaticPropagationSolution.plot` otherwise.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        axs : axes
            The axis handles.
        """
        nodes = self.path.nodes_only
        # From node is always first node of cavity path
        fn = nodes[0]
        traversed_ports = set()
        tn = None
        for node in nodes:  # traverse all the nodes
            # If port has already been encountered, stop
            # as we don't want to repeat the same space
            # (this is applicable to linear cavs only)
            if node.port in traversed_ports:
                tn = node
                break

            traversed_ports.add(node.port)

        # If we didn't traverse the same port twice (e.g. this
        # is a ring or bow-tie type cavity) then set to node as
        # final node of the cavity path for the round-trip
        if tn is None:
            tn = nodes[-1]

        if direction is None:
            ps = self._model.propagate_beam_astig(from_node=fn, to_node=tn)
        else:
            ps = self._model.propagate_beam(
                from_node=fn, to_node=tn, direction=direction
            )

        return ps.plot(*args, **kwargs)



import pytest
from ..data import SI_PREFICES, INTEGERS, FLOATS, FLOATS_STD, IMAGINARIES
from ..util import (
    NAME,
    NUMBER,
    MINUS,
    LPAREN,
    RPAREN,
    EQUALS,
    SPACE,
    IMPLICITLINEEND,
    ENDMARKER,
)


@pytest.mark.parametrize(
    "string,expected",
    (
        ("12345", [NUMBER(1, 1, "12345"), IMPLICITLINEEND(1, 6), ENDMARKER(2),]),
        (
            "(-124561-1)",
            [
                LPAREN(1, 1),
                MINUS(1, 2),
                NUMBER(1, 3, "124561"),
                MINUS(1, 9),
                NUMBER(1, 10, "1"),
                RPAREN(1, 11),
                IMPLICITLINEEND(1, 12),
                ENDMARKER(2),
            ],
        ),
        # Long integer.
        (
            "a = 123141242151251616110",
            [
                NAME(1, 1, "a"),
                SPACE(1, 2),
                EQUALS(1, 3),
                SPACE(1, 4),
                NUMBER(1, 5, "123141242151251616110"),
                IMPLICITLINEEND(1, 26),
                ENDMARKER(2),
            ],
        ),
    ),
)
def test_int(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected


@pytest.mark.parametrize(
    "string,expected",
    (
        ("3.14159", [NUMBER(1, 1, "3.14159"), IMPLICITLINEEND(1, 8), ENDMARKER(2),]),
        ("314159.", [NUMBER(1, 1, "314159."), IMPLICITLINEEND(1, 8), ENDMARKER(2),]),
        (".314159", [NUMBER(1, 1, ".314159"), IMPLICITLINEEND(1, 8), ENDMARKER(2),]),
        ("3e14159", [NUMBER(1, 1, "3e14159"), IMPLICITLINEEND(1, 8), ENDMARKER(2),]),
        ("3E123", [NUMBER(1, 1, "3E123"), IMPLICITLINEEND(1, 6), ENDMARKER(2),]),
        ("3e-1230", [NUMBER(1, 1, "3e-1230"), IMPLICITLINEEND(1, 8), ENDMARKER(2),]),
        ("3.14e159", [NUMBER(1, 1, "3.14e159"), IMPLICITLINEEND(1, 9), ENDMARKER(2),]),
    ),
)
def test_float(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected


@pytest.mark.parametrize(
    "string,expected",
    (
        ("3.14159j", [NUMBER(1, 1, "3.14159j"), IMPLICITLINEEND(1, 9), ENDMARKER(2),]),
        ("314159.j", [NUMBER(1, 1, "314159.j"), IMPLICITLINEEND(1, 9), ENDMARKER(2),]),
        (".314159j", [NUMBER(1, 1, ".314159j"), IMPLICITLINEEND(1, 9), ENDMARKER(2),]),
        ("3e14159j", [NUMBER(1, 1, "3e14159j"), IMPLICITLINEEND(1, 9), ENDMARKER(2),]),
        ("3E123j", [NUMBER(1, 1, "3E123j"), IMPLICITLINEEND(1, 7), ENDMARKER(2),]),
        ("3e-1230j", [NUMBER(1, 1, "3e-1230j"), IMPLICITLINEEND(1, 9), ENDMARKER(2),]),
        (
            "3.14e159j",
            [NUMBER(1, 1, "3.14e159j"), IMPLICITLINEEND(1, 10), ENDMARKER(2),],
        ),
    ),
)
def test_imaginary(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected


@pytest.mark.parametrize("number,_a,_b", INTEGERS + FLOATS + IMAGINARIES)
def test_numbers(tokenizer, number, _a, _b):
    assert list(tokenizer.tokenize(number)) == [
        NUMBER(1, 1, number),
        IMPLICITLINEEND(1, 1 + len(str(number))),
        ENDMARKER(2),
    ]


@pytest.mark.parametrize("number,_a,_b", INTEGERS + FLOATS_STD)
@pytest.mark.parametrize("prefix,exponent", SI_PREFICES.items())
def test_numbers_with_si_prefices(tokenizer, number, _a, _b, prefix, exponent):
    string = f"{number}{prefix}"
    assert list(tokenizer.tokenize(string)) == [
        NUMBER(1, 1, string),
        IMPLICITLINEEND(1, 1 + len(str(string))),
        ENDMARKER(2),
    ]



"""Tests for making sure that order in katscript doesn't matter."""

import pytest
from hypothesis import given, settings
from hypothesis.strategies import permutations
from finesse.script import parse
from .....diff import assert_models_equivalent


# A script with a `cav` command which normally has to be built last.
SCRIPT = """
l L0 P=1
s s0 L0.p1 EOM1.p1
mod EOM1 f=100M midx=0.1 order=1 mod_type=pm
s s1 EOM1.p2 ITM.p1 L=1
m ITM R=0.99 T=0.01
s sCAV ITM.p2 ETM.p1 L=1
m ETM R=0.99 T=0.01 phi=123
pd1 REFL_I ITM.p1.o 100M 0
cav FP ITM.p2.o
"""


@pytest.fixture(scope="package")
def reference_model():
    """The parsed model without any re-ordering."""
    return parse(SCRIPT)


@given(scriptlines=permutations(SCRIPT.splitlines()))
@settings(deadline=1000)  # See #292.
def test_script_order_doesnt_matter(scriptlines, reference_model):
    """Test that script order does not matter.

    This test works by randomly reordering the lines of the reference script, then
    parsing and comparing the model to the reference model.
    """
    model = parse("\n".join(scriptlines))
    assert_models_equivalent(reference_model, model)



"""Kat file parser mode selection syntax tests."""

import re
import pytest
import numpy as np
from finesse.script import parse
from finesse.script.exceptions import KatScriptError


@pytest.mark.parametrize(
    "modes,expected",
    (
        # Keywords.
        ("even, 4", [[0, 0], [0, 2], [0, 4], [2, 0], [2, 2], [4, 0]]),
        ("odd, 3", [[0, 0], [0, 1], [0, 3], [1, 0], [1, 1], [3, 0]]),
        ("x, 5", [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0], [5, 0]]),
        ("y, 4", [[0, 0], [0, 1], [0, 2], [0, 3], [0, 4]]),
        # Arrays.
        ("[[0, 0], [1, 1], [2, 2], [1, 1], [0, 0]]", [[0, 0], [1, 1], [2, 2]]),
    ),
)
def test_modes(modes, expected):
    model = parse(f"modes({modes})")
    assert np.all(model.homs == expected)


@pytest.mark.parametrize(
    "modes,error",
    (
        pytest.param(
            "[3, 2]",
            "\nline 1: Expected mode list to be a two-dimensional list\n"
            "-->1: modes([3, 2])\n"
            "      ^^^^^",
            id="[3, 2]",
        ),
        pytest.param(
            "[[[0, 1], [1, 0]], [1, 2], [3, 4], [5, 6]]",
            "\nline 1: Expected n (= [0, 1]) and m (= [1, 0]) of element [[0, 1], [1, 0]] of mode "
            "list to be convertible to non-negative integers.\n"
            "-->1: modes([[[0, 1], [1, 0]], [1, 2], [3, 4], [5, 6]])\n"
            "      ^^^^^",
            id=">2D nested list",
        ),
        pytest.param(
            "maxtem=-1",
            "\nline 1: Argument maxtem must be a non-negative integer.\n"
            "-->1: modes(maxtem=-1)\n"
            "      ^^^^^",
            id="negative integer",
        ),
        pytest.param(
            "maxtem=3.141",
            "\nline 1: Argument maxtem must be a non-negative integer.\n"
            "-->1: modes(maxtem=3.141)\n"
            "      ^^^^^",
            id="non integer",
        ),
    ),
)
def test_invalid_modes(modes, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        parse(f"modes({modes})")



"""Cavity tests."""

import pytest
import numpy as np
from finesse import Model
from finesse.components import Cavity, Mirror, Beamsplitter


@pytest.fixture(scope="module")
def equivalent_cavities():
    """Fixture to provide different types of cavity (2-, 3-, 4-mirror) that are equivalent."""
    T1 = 500e-6
    T2 = 30e-6
    # Equivalent of T2 through 2 mirrors.
    T2T = 1 - (1 - T2) ** (1 / 2)
    # Equivalent of T2 through 3 mirrors.
    T2Q = 1 - (1 - T2) ** (1 / 3)

    # 10 m linear cavity model.
    m11 = Mirror("m11", T=T1, L=0, Rc=12)
    m12 = Mirror("m12", T=T2, L=0)
    model_linear = Model()
    model_linear.connect(m11.p1, m12.p1, L=9)

    # 10 m triple cavity model.
    # Needs to use beam splitters because the incoming and outgoing beams are separate.
    m21 = Beamsplitter("m21", T=T1, L=0, Rc=12)
    m22 = Beamsplitter("m22", T=T2T, L=0)
    m23 = Beamsplitter("m23", T=T2T, L=0)
    model_triple = Model()
    model_triple.connect(m21.p1, m22.p2, L=6)
    model_triple.connect(m22.p1, m23.p2, L=6)
    model_triple.connect(m23.p1, m21.p2, L=6)

    # 10 m quadruple cavity model.
    # Needs to use beam splitters because the incoming and outgoing beams are separate.
    m31 = Beamsplitter("m31", T=T1, L=0, Rc=12)
    m32 = Beamsplitter("m32", T=T2Q, L=0)
    m33 = Beamsplitter("m33", T=T2Q, L=0)
    m34 = Beamsplitter("m34", T=T2Q, L=0)
    model_quad = Model()
    model_quad.connect(m31.p1, m32.p2, L=4.5)
    model_quad.connect(m32.p1, m33.p2, L=4.5)
    model_quad.connect(m33.p1, m34.p2, L=4.5)
    model_quad.connect(m34.p1, m31.p2, L=4.5)

    # Cavity models.
    cav_linear = Cavity("cav_linear", m11.p1.o)
    cav_triple = Cavity("cav_triple", m21.p1.o, m22.p2.i)
    cav_quad = Cavity("cav_quad", m31.p1.o, m32.p2.i)

    model_linear.add(cav_linear)
    model_triple.add(cav_triple)
    model_quad.add(cav_quad)

    # Need to return the models here too because they're only contained inside the cavities as
    # weakrefs and could be lost to garbage collection otherwise.
    return (
        (model_linear, cav_linear),
        (model_triple, cav_triple),
        (model_quad, cav_quad),
    )


def test_equivalent_cavities_round_trip_length(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same round trip \
    length."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities
    assert (
        cav_linear.round_trip_length
        == cav_triple.round_trip_length
        == cav_quad.round_trip_length
        != 0
    )


def test_equivalent_cavities_abcd(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same ABCD matrices.

    Because only one mirror in each cavity has a radius of curvature, and the round-trip lengths
    are the same, the cavities should have the same ABCD matrices as long as the same surfaces
    are connected (for the Rc sign convention). However, because the triple cavity has an odd
    number of reflections, the x-plane ABCD matrix is flipped.
    """
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities
    assert np.all(cav_linear.ABCDx == cav_quad.ABCDx)
    assert np.all(cav_linear.ABCDx == -cav_triple.ABCDx)
    assert np.all(cav_linear.ABCDy == cav_quad.ABCDy)
    assert np.all(cav_linear.ABCDy == cav_triple.ABCDy)


def test_equivalent_cavities_gouy(equivalent_cavities):
    """Test that 2- and 4-mirror cavities with equivalent properties have same gouy phase, but \
    that the 3-mirror equivalent does not.

    The triple cavity has an odd number of reflections, so the gouy phase in the x-direction is
    different.
    """
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.gouy_x != 0
    assert cav_linear.gouy_x != cav_triple.gouy_x
    assert cav_linear.gouy_x == cav_quad.gouy_x

    assert cav_linear.gouy_y != 0
    assert cav_linear.gouy_y == cav_triple.gouy_y
    assert cav_linear.gouy_y == cav_quad.gouy_y


def test_equivalent_cavities_power_and_loss(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same power and \
    loss."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.loss != 0
    assert cav_triple.loss != 0
    assert cav_quad.loss != 0
    assert np.isclose(cav_linear.loss, cav_triple.loss)
    assert np.isclose(cav_linear.loss, cav_quad.loss)


def test_equivalent_cavities_finesse(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same finesse."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.finesse != 0
    assert cav_triple.finesse != 0
    assert cav_quad.finesse != 0
    assert np.isclose(cav_linear.finesse, cav_triple.finesse)
    assert np.isclose(cav_linear.finesse, cav_quad.finesse)


def test_equivalent_cavities_fsr(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same FSR."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.FSR != 0
    assert cav_triple.FSR != 0
    assert cav_quad.FSR != 0
    assert cav_linear.FSR == cav_triple.FSR
    assert cav_linear.FSR == cav_quad.FSR


def test_equivalent_cavities_fwhm(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same FWHM."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.FWHM != 0
    assert cav_triple.FWHM != 0
    assert cav_quad.FWHM != 0
    assert np.isclose(cav_linear.FWHM, cav_triple.FWHM)
    assert np.isclose(cav_linear.FWHM, cav_quad.FWHM)


def test_equivalent_cavities_pole(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same pole."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.pole != 0
    assert cav_triple.pole != 0
    assert cav_quad.pole != 0
    assert np.isclose(cav_linear.pole, cav_triple.pole)
    assert np.isclose(cav_linear.pole, cav_quad.pole)


def test_equivalent_cavities_mode_separation(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same mode \
    separation.

    The triple cavity has an odd number of reflections, so the mode separation in the x-direction is
    different.
    """
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.mode_separation_x != 0
    assert cav_triple.mode_separation_x != 0
    assert cav_quad.mode_separation_x != 0
    assert cav_linear.mode_separation_x != cav_triple.mode_separation_x
    assert np.isclose(cav_linear.mode_separation_x, cav_quad.mode_separation_x)

    assert cav_linear.mode_separation_y != 0
    assert cav_triple.mode_separation_y != 0
    assert cav_quad.mode_separation_y != 0
    assert np.isclose(cav_linear.mode_separation_y, cav_triple.mode_separation_y)
    assert np.isclose(cav_linear.mode_separation_y, cav_quad.mode_separation_y)


def test_equivalent_cavities_q(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same q."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.qx != 0
    assert cav_triple.qx != 0
    assert cav_quad.qx != 0
    assert cav_linear.qx == cav_triple.qx
    assert cav_linear.qx == cav_quad.qx

    assert cav_linear.qy != 0
    assert cav_triple.qy != 0
    assert cav_quad.qy != 0
    assert cav_linear.qy == cav_triple.qy
    assert cav_linear.qy == cav_quad.qy


def test_equivalent_cavities_resolution(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same resolution.

    The triple cavity has an odd number of reflections, so the resolution in the x-direction is
    different.
    """
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.Sx != 0
    assert cav_triple.Sx != 0
    assert cav_quad.Sx != 0
    assert cav_linear.Sx != cav_triple.Sx
    assert np.isclose(cav_linear.Sx, cav_quad.Sx)

    assert cav_linear.Sy != 0
    assert cav_triple.Sy != 0
    assert cav_quad.Sy != 0
    assert np.isclose(cav_linear.Sy, cav_triple.Sy)
    assert np.isclose(cav_linear.Sy, cav_quad.Sy)


def test_equivalent_cavities_g(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties have same g (stability).

    The triple cavity has an odd number of reflections, so the resolution in the x-direction is
    different.
    """
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert cav_linear.gx != 0
    assert cav_triple.gx != 0
    assert cav_quad.gx != 0
    assert cav_linear.gx != cav_triple.gx
    assert cav_linear.gx == cav_quad.gx

    assert cav_linear.gy != 0
    assert cav_triple.gy != 0
    assert cav_quad.gy != 0
    assert cav_linear.gy == cav_triple.gy
    assert cav_linear.gy == cav_quad.gy


def test_equivalent_cavities_stable(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties are equally stable."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert all((cav_linear.is_stable_x, cav_triple.is_stable_x, cav_quad.is_stable_x))
    assert all((cav_linear.is_stable_y, cav_triple.is_stable_y, cav_quad.is_stable_y))


def test_equivalent_cavities_critical(equivalent_cavities):
    """Test that 2-, 3- and 4-mirror cavities with equivalent properties are equally (not) \
    critical."""
    (_, cav_linear), (_, cav_triple), (_, cav_quad) = equivalent_cavities

    assert not any(
        (cav_linear.is_critical_x, cav_triple.is_critical_x, cav_quad.is_critical_x)
    )
    assert not any(
        (cav_linear.is_critical_x, cav_triple.is_critical_x, cav_quad.is_critical_x)
    )



"""
Squeezer-type optical components for producing squeezed light inputs.
"""
from finesse.parameter import float_parameter

from finesse.components.general import (
    Connector,
    FrequencyGenerator,
    NoiseGenerator,
    NoiseType,
)
from finesse.components.modal.squeezer import (
    squeezer_fill_rhs,
    squeezer_fill_qnoise,
    SqueezerWorkspace,
)
from finesse.components.node import NodeType, NodeDirection


@float_parameter("db", "dB", units="dB")
@float_parameter(
    "f", "Frequency", units="Hz"
)
@float_parameter("angle", "Angle", units="degrees")
class Squeezer(Connector, FrequencyGenerator, NoiseGenerator):
    """
    Represents a squeezer producing a squeezed-light beam with
    a given squeezing factor and angle.

    Parameters
    ----------
    name : str
        Name of the newly created squeezer.

    db : float
        Squeezing factor (in decibels).

    f : float or :class:`.Frequency`, optional
        Frequency-offset of the squeezer from the default (in Hz) or
        :class:`.Frequency` object. Defaults to 0 Hz offset.

    angle : float, optional
        Squeezing angle (in degrees). Defaults to zero.
    """

    def __init__(self, name, db, f=0, angle=0):
        Connector.__init__(self, name)
        FrequencyGenerator.__init__(self)
        NoiseGenerator.__init__(self)

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self.db = db
        self.f = f
        self.angle = angle

        self._register_noise_output("P1o", self.p1.o, NoiseType.QUANTUM)

    def _source_frequencies(self):
        return [self.f.ref]

    def _couples_noise(self, ws, node, noise_type, frequency_in, frequency_out):
        return (frequency_in.index == frequency_out.index) or (
            (frequency_in.audio_carrier_index == frequency_out.audio_carrier_index)
            and (frequency_in.audio_carrier_index == ws.fsrc_car_idx)
        )

    def _get_workspace(self, sim):
        ws = SqueezerWorkspace(self, sim, True)
        ws.fsrc_car_idx = -1

        if not sim.signal:
            ws.node_id = sim.carrier.node_id(self.p1.o)
            # Carrier just fills RHS
            ws.set_fill_rhs_fn(squeezer_fill_rhs)
            fsrc = self.__find_src_freq(sim)
            # Didn't find a Frequency bin for this squeezer in carrier simulation
            if fsrc is None:
                raise Exception(
                    f"Could not find a frequency bin at {self.f} for {self}"
                )
            ws.fsrc_car_idx = fsrc.index
        else:
            ws.node_id = sim.signal.node_id(self.p1.o)
            ws.signal.set_fill_noise_function(NoiseType.QUANTUM, squeezer_fill_qnoise)
            fsrc = self.__find_src_freq(sim.carrier)
            # Didn't find a Frequency bin for this squeezer in carrier simulation
            if fsrc is None:
                raise Exception(
                    f"Could not find a frequency bin in carrier sim at {self.f} for {self}"
                )
            ws.fsrc_car_idx = fsrc.index

        # if sim.is_modal: self._update_tem_gouy_phases(sim)
        return ws

    def __find_src_freq(self, sim):
        # if it's tunable we want to look for the symbol that is just this
        # lasers frequency, as it will be changing
        for f in sim.optical_frequencies.frequencies:
            if not self.f.is_changing:
                # Don't match changing frequency bins if ours won't match
                if not f.symbol.is_changing and (
                    f.f == self.f.value  # match potential param refs
                    or f.f == float(self.f.value)  # match numeric values
                ):
                    # If nothing is changing then we can just match freq values
                    return f
            else:
                # If our frequency is changing then we have to have a frequency bin that
                # matches our symbol
                if f.symbol == self.f.ref:
                    return f  # Simple case
        return None



import re
import pytest
from finesse.element import ModelElement
from finesse.parameter import float_parameter
from finesse.script.exceptions import KatScriptError
from ....util import dedent_multiline


@float_parameter("fake_param", "Fake parameter")
class FakeElement(ModelElement):
    def __init__(self, name, fake_param=0.0):
        super().__init__(name)
        self.fake_param = fake_param


@pytest.fixture
def compiler(compiler, fake_binop_cls):
    compiler.spec.register_element(
        "fake_element", {"setter": FakeElement, "getter": FakeElement}
    )
    compiler.spec.binary_operators["-"] = lambda a, b: fake_binop_cls("-", a, b)

    return compiler


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 &myelement2.fake_param
                fake_element myelement2 &myelement1.fake_param
                """
            ),
            (
                "\nlines 1-2: cyclic parameters\n"
                "-->1: fake_element myelement1 &myelement2.fake_param\n"
                "                              ^^^^^^^^^^^^^^^^^^^^^^\n"
                "-->2: fake_element myelement2 &myelement1.fake_param\n"
                "                              ^^^^^^^^^^^^^^^^^^^^^^"
            ),
            id="cycle between two elements",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 &myelement2.fake_param
                fake_element myelement2 1-&myelement1.fake_param
                """
            ),
            (
                "\nlines 1-2: cyclic parameters\n"
                "-->1: fake_element myelement1 &myelement2.fake_param\n"
                "                              ^^^^^^^^^^^^^^^^^^^^^^\n"
                "-->2: fake_element myelement2 1-&myelement1.fake_param\n"
                "                                ^^^^^^^^^^^^^^^^^^^^^^"
            ),
            id="cycle between two elements (one inside expression)",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 &myelement2.fake_param
                fake_element myelement2 -&myelement1.fake_param
                """
            ),
            (
                "\nlines 1-2: cyclic parameters\n"
                "-->1: fake_element myelement1 &myelement2.fake_param\n"
                "                              ^^^^^^^^^^^^^^^^^^^^^^\n"
                "-->2: fake_element myelement2 -&myelement1.fake_param\n"
                "                               ^^^^^^^^^^^^^^^^^^^^^^"
            ),
            id="cycle between two elements (one inside unary function)",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 &myelement2.fake_param
                fake_element myelement2 &myelement3.fake_param
                fake_element myelement3 &myelement4.fake_param
                fake_element myelement4 &myelement1.fake_param
                """
            ),
            (
                "\nlines 1-4: cyclic parameters\n"
                "-->1: fake_element myelement1 &myelement2.fake_param\n"
                "                              ^^^^^^^^^^^^^^^^^^^^^^\n"
                "-->2: fake_element myelement2 &myelement3.fake_param\n"
                "                              ^^^^^^^^^^^^^^^^^^^^^^\n"
                "-->3: fake_element myelement3 &myelement4.fake_param\n"
                "                              ^^^^^^^^^^^^^^^^^^^^^^\n"
                "-->4: fake_element myelement4 &myelement1.fake_param\n"
                "                              ^^^^^^^^^^^^^^^^^^^^^^"
            ),
            id="cycle between four elements",
        ),
    ),
)
def test_cycles_invalid(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)



import logging
import pytest
from finesse.utilities import logs


# NOTE: pytest caplog seems to ignore the configured log format, so setting `fmt` and
# `datefmt` below has no effect.
@pytest.mark.parametrize(
    "context_level", ("debug", "info", "warning", "error", "critical")
)
@pytest.mark.parametrize(
    "message_level", ("debug", "info", "warning", "error", "critical")
)
@pytest.mark.parametrize("channel", ("finesse", "finesse.mod1", "finesse.mod1.mod2"))
def test_log_context(caplog, context_level, message_level, channel):
    """Test log context manager.

    Notes
    -----
    Pytest's caplog fixture uses its own handler and so the channel exclude and
    formatting features of :func:`.logs` can't be tested.
    """
    message_level = logging.getLevelName(message_level.upper())
    context_level = logging.getLevelName(context_level.upper())

    with logs(level=context_level):
        logger = logging.getLogger(channel)
        logger.log(message_level, "__test__")

        if message_level >= context_level:
            assert len(caplog.records) == 1
            assert caplog.records[0].getMessage() == "__test__"
        else:
            assert len(caplog.records) == 0



"""Tests for mode masks on detectors."""

import logging
import numpy as np
import pytest
from finesse import Model
from finesse.analysis import noxaxis
from finesse.utilities.homs import make_modes, remove_modes


@pytest.fixture
def fp_cavity_model():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    s s0 L0.p1 ITM.p1

    m ITM R=0.9 T=0.1 Rc=-2
    s sc ITM.p2 ETM.p1 L=1
    m ETM R=0.9 T=0.1 Rc=2

    pd P_total ETM.p2.o
    pd P_masked ETM.p2.o

    cav FP ITM.p2
    """
    )
    IFO.ITM.Rcy = IFO.ITM.Rcx.ref
    IFO.ETM.Rcy = IFO.ETM.Rcx.ref

    return IFO


def test_masked_tem00_gives_null(fp_cavity_model: Model):
    """Test that masking the 00 mode for a model with no mismatches nor misalignments
    gives zero response in the masked power detector."""
    IFO = fp_cavity_model

    IFO.select_modes("even", maxtem=4)

    # Mask the 00 mode
    IFO.P_masked.mask = ["00"]

    # Check the mask itself has been interpreted correctly
    assert np.all(IFO.P_masked.mask == [[0, 0]])

    out = noxaxis(IFO)

    # No mismatches nor misalignments present so masking
    # HG00 should give a null response
    assert out["P_masked"] == 0
    # Impedance matched so total transmitted power should
    # equal input power
    assert out["P_total"] == pytest.approx(IFO.L0.P.value)


def test_masked_tem00_gives_null_ccd_image(fp_cavity_model: Model):
    """Test that masking the 00 mode for a model with no mismatches nor misalignments
    gives zero response in the masked CCD."""
    IFO = fp_cavity_model

    Npts = 100
    IFO.select_modes("even", maxtem=4)
    IFO.parse(f"ccd CCD ETM.p2.o xlim=5 ylim=5 npts={Npts}")

    # Mask the 00 mode
    IFO.CCD.mask = ["00"]

    # Check the mask itself has been interpreted correctly
    assert np.all(IFO.CCD.mask == [[0, 0]])

    out = noxaxis(IFO)

    # No mismatches nor misalignments present so masking
    # HG00 should give an image of zeros
    assert np.all(out["CCD"] == np.zeros((Npts, Npts)))


def test_masked_tem00_gives_null_field_image(fp_cavity_model: Model):
    """Test that masking the 00 mode for a model with no mismatches nor misalignments
    gives zero response in the masked FieldCamera."""
    IFO = fp_cavity_model

    Npts = 100
    IFO.select_modes("even", maxtem=4)
    IFO.parse(f"fcam FCAM ETM.p2.o f=0 xlim=5 ylim=5 npts={Npts}")

    # Mask the 00 mode
    IFO.FCAM.mask = ["00"]

    # Check the mask itself has been interpreted correctly
    assert np.all(IFO.FCAM.mask == [[0, 0]])

    out = noxaxis(IFO)

    # No mismatches nor misalignments present so masking
    # HG00 should give an image of complex zeros
    assert np.all(out["FCAM"] == np.zeros((Npts, Npts), dtype=np.complex128))


def test_masked_even_order_tem_mismatched_file_gives_null(fp_cavity_model: Model):
    """Test that masking all even-order modes for a mismatched model gives zero response
    in the masked power detector."""
    IFO = fp_cavity_model

    IFO.select_modes("even", maxtem=4)
    # Add another mode so that size of P_masked.mask is
    # not same as number of HOMs in the model
    IFO.include_modes("01")

    # Create mismatch originating at L0 of 5% in w0, 5% in z
    IFO.create_mismatch(IFO.L0.p1.o, w0_mm=5, z_mm=5)

    # Mask all even order modes
    IFO.P_masked.select_mask("even")

    assert np.all(IFO.P_masked.mask == make_modes("even", maxtem=4))

    out = noxaxis(IFO)

    # Only mismatch present (no misalignment) so masking all
    # even order modes on P_masked should give a null response
    assert out["P_masked"] == 0


def test_masked_x_modes_itmyaw_file_gives_null(fp_cavity_model: Model):
    """Test that masking all x modes for a model with yawed mirror gives zero response
    in the masked power detector."""
    IFO = fp_cavity_model

    # Apply a yaw misalignment to ITM of 1% of the cavity divergence angle
    IFO.ITM.xbeta = IFO.FP.qx.divergence / 100

    IFO.select_modes("x", maxtem=4)
    # Add another mode so that size of P_masked.mask is
    # not same as number of HOMs in the model
    IFO.include_modes("02")

    # Mask all x order modes
    IFO.P_masked.select_mask("x")

    assert np.all(IFO.P_masked.mask == make_modes("x", maxtem=4))

    out = noxaxis(IFO)

    # Only yaw misalignment present so masking all yaw modes
    # on P_masked should give a null response
    assert out["P_masked"] == 0


def test_masked_y_modes_itmpitch_file_gives_null(fp_cavity_model: Model):
    """Test that masking all x modes for a model with yawed mirror gives zero response
    in the masked power detector."""
    IFO = fp_cavity_model

    # Apply a pitch misalignment to ITM of 1% of the cavity divergence angle
    IFO.ITM.ybeta = IFO.FP.qx.divergence / 100

    IFO.select_modes("y", maxtem=4)
    # Add another mode so that size of P_masked.mask is
    # not same as number of HOMs in the model
    IFO.include_modes("20")

    # Mask all x order modes
    IFO.P_masked.select_mask("y")

    assert np.all(IFO.P_masked.mask == make_modes("y", maxtem=4))

    out = noxaxis(IFO)

    # Only pitch misalignment present so masking all pitch modes
    # on P_masked should give a null response
    assert out["P_masked"] == 0


def test_masked_tem00_mismatched_file(fp_cavity_model: Model):
    """Test that masking the 00 mode for a mismatched model gives response equal to abs
    sum sqd of HOM field amplitudes."""
    IFO = fp_cavity_model

    IFO.select_modes("even", maxtem=4)
    for n, m in IFO.homs:
        IFO.parse(f"ad ad{n}{m} ETM.p2.o 0 n={n} m={m}")

    # Create mismatch originating at L0 of 5% in w0, 5% in z
    IFO.create_mismatch(IFO.L0.p1.o, w0_mm=5, z_mm=5)

    # Mask the 00 mode
    IFO.P_masked.mask = ["00"]
    IFO.ad00.mask = ["00"]

    assert np.all(IFO.P_masked.mask == [[0, 0]])
    assert np.all(IFO.ad00.mask == [[0, 0]])

    out = noxaxis(IFO)
    hom_amp_sum = 0
    for n, m in IFO.homs:
        if not n and not m:
            continue

        hom_amp_sum += np.abs(out[f"ad{n}{m}"] ** 2)

    # Only mismatch present so masking 00 mode means
    # P_masked should equal abs sum sqd of HOm fields
    assert out["P_masked"] == pytest.approx(hom_amp_sum)

    # Masking the mode itself on an ad should always give zero
    assert out["ad00"] == 0 + 0j


def test_masked_even_modes_mismatched_file_detect_tem00(fp_cavity_model: Model):
    """Test that masking all modes but 00 for a mismatched model gives response equal to
    abs sqd of a00 field."""
    IFO = fp_cavity_model

    IFO.select_modes("even", maxtem=4)
    IFO.parse("ad ad00 ETM.p2.o 0 n=0 m=0")

    # Create mismatch originating at L0 of 5% in w0, 5% in z
    IFO.create_mismatch(IFO.L0.p1.o, w0_mm=5, z_mm=5)

    # Mask all but 00 mode
    IFO.P_masked.select_mask("even", exclude="00")

    mask = make_modes("even", maxtem=4)
    mask = remove_modes(mask, "00")
    assert np.all(IFO.P_masked.mask == mask)

    out = noxaxis(IFO)

    # Only mismatch present so masking even modes means
    # P_masked should equal abs sqd of 00 field
    assert out["P_masked"] == pytest.approx(np.abs(out["ad00"]) ** 2)


def test_masking_all_modes_raises_exception(fp_cavity_model: Model):
    """Test that masking all modes on a detector raises a RuntimeError."""
    IFO = fp_cavity_model

    IFO.select_modes("even", maxtem=4)

    # Mask all modes...
    IFO.P_masked.select_mask("even")

    # ... and check that this raises an error
    with pytest.raises(RuntimeError):
        noxaxis(IFO)


def test_masking_non_existent_mode_raises_exception(fp_cavity_model: Model):
    """Test that masking modes which are not present in the model raises a
    RuntimeError."""
    IFO = fp_cavity_model

    IFO.select_modes("even", maxtem=4)

    # Mask modes which aren't in the model...
    IFO.P_masked.select_mask(["01", "33"])

    # ... and check that this raises an error
    with pytest.raises(RuntimeError):
        noxaxis(IFO)


def test_masking_ad_single_field_mode_logs_warning(fp_cavity_model: Model, caplog):
    """Test that applying mask to the mode that an amplitude detector is detecting logs
    a warning and outputs complex zero."""
    IFO = fp_cavity_model

    IFO.select_modes(maxtem=2)
    IFO.parse("ad ad00 ETM.p2.o f=0 n=0 m=0")

    IFO.ad00.mask = ["00"]

    with caplog.at_level(logging.WARNING):
        out = noxaxis(IFO)

        assert "This will always return values of complex zero" in caplog.text

    assert out["ad00"] == complex(0, 0)


def test_masking_ad_single_field_mode_not_detected_logs_warning(
    fp_cavity_model: Model, caplog
):
    """Test that applying mask to the mode that an amplitude detector is not detecting
    logs a warning."""
    IFO = fp_cavity_model

    IFO.select_modes(maxtem=2)
    IFO.parse("ad ad00 ETM.p2.o f=0 n=0 m=0")

    IFO.ad00.mask = ["01"]

    with caplog.at_level(logging.WARNING):
        noxaxis(IFO)

        assert "Mask applied to ad00 has no effect" in caplog.text


# TODO (sjr) Above covers masks on pd0, CCD, ad single field. Need to
#            add mode mask tests for:
#                - ad in multi-field mode (i.e. n=m=None)
#                - pd1 and pd2
